"""Copyright (c) 2015 Red Hat, Inc
All rights reserved.

This software may be modified and distributed under the terms
of the BSD license. See the LICENSE file for details.

Sync built image to pulp registry using Docker Registry HTTP API V2

Authentication is via a key and certificate in a secret which the
builder service account is allowed to mount:

$ oc secrets new pulp ./pulp.key ./pulp.cer
secrets/pulp
$ oc secrets add serviceaccount/builder secret/pulp --for=mount

In the BuildConfig for atomic-reactor, specify the secret in the
strategy's 'secrets' array, specifying a mount path:

"secrets": [{
  "secretSource": {
    "name": "pulp"
  },
  "mountPath": "/var/run/secrets/pulp"
}]

In the configuration for this plugin, specify the same path for
pulp_secret_path:

"pulp_sync": {
  "pulp_registry_name": ...,
  ...
  "pulp_secret_path": "/var/run/secrets/pulp"
}

"""

from __future__ import print_function, unicode_literals

from atomic_reactor.plugin import PostBuildPlugin
from atomic_reactor.util import ImageName
from contextlib import contextmanager
import dockpulp
import os
import re
from tempfile import NamedTemporaryFile


# let's silence warnings from dockpulp: there is one warning for every
# request which may result in tenths of messages: very annoying with
# "module", it just prints one warning -- this should balance security
# and UX
from warnings import filterwarnings
filterwarnings("module")


@contextmanager
def dockpulp_config(docker_registry, **kwargs):
    """
    Temporary dockpulp config pointing to the docker v2 registry

    :yields: NamedTemporaryFile instance
    """

    env = 'sync'
    template = """
[registries]
{sync}

[filers]
{sync}

[pulps]
{sync}
"""
    sync = '{env} = {url}'.format(env=env, url=docker_registry)
    with NamedTemporaryFile('wt', **kwargs) as config:
        config.write(template.format(sync=sync))
        config.flush()
        config.env = env
        yield config


class PulpSyncPlugin(PostBuildPlugin):
    key = "pulp_sync"
    is_allowed_to_fail = False

    CER = 'pulp.cer'
    KEY = 'pulp.key'

    def __init__(self, tasker, workflow,
                 pulp_registry_name,
                 docker_registry,
                 delete_from_registry=False,
                 pulp_secret_path=None,
                 username=None, password=None,
                 dockpulp_loglevel=None):
        """
        constructor

        :param tasker: DockerTasker instance
        :param workflow: DockerBuildWorkflow instance
        :param pulp_registry_name: str, name of pulp registry to use,
               specified in /etc/dockpulp.conf
        :param docker_registry: str, URL of docker registry to sync from
        :param delete_from_registry: bool, whether to delete the image
               from the docker v2 registry after sync
        :param pulp_secret_path: path to pulp.cer and pulp.key
        :param username: pulp username, used in preference to
               certificate and key
        :param password: pulp password, used in preference to
               certificate and key
        """
        # call parent constructor
        super(PulpSyncPlugin, self).__init__(tasker, workflow)
        self.pulp_registry_name = pulp_registry_name
        self.docker_registry = docker_registry
        self.pulp_secret_path = pulp_secret_path
        self.username = username
        self.password = password

        if dockpulp_loglevel is not None:
            logger = dockpulp.setup_logger(dockpulp.log)
            try:
                logger.setLevel(dockpulp_loglevel)
            except (ValueError, TypeError) as ex:
                self.log.error("Can't set provided log level %r: %r",
                               dockpulp_loglevel, ex)

        if delete_from_registry:
            self.log.error("will not delete from registry as instructed: "
                           "not implemented")

    def set_auth(self, pulp):
        if self.username and self.password:
            # Use username and password if provided
            pulp.login(self.username, self.password)
        elif self.pulp_secret_path or 'SOURCE_SECRET_PATH' in os.environ:
            if self.pulp_secret_path is not None:
                path = self.pulp_secret_path
                self.log.info("using configured path %s for secrets", path)
            else:
                path = os.environ["SOURCE_SECRET_PATH"]
                self.log.info("SOURCE_SECRET_PATH=%s from environment", path)

            # Work out the pathnames for the certificate/key pair
            cer = os.path.join(path, self.CER)
            key = os.path.join(path, self.KEY)

            if not os.path.exists(cer):
                raise RuntimeError("Certificate does not exist")
            if not os.path.exists(key):
                raise RuntimeError("Key does not exist")

            # Tell dockpulp
            pulp.set_certs(cer, key)

    def run(self):
        pulp = dockpulp.Pulp(env=self.pulp_registry_name)
        self.set_auth(pulp)

        # We only want the hostname[:port]
        pulp_registry = re.sub(r'^https?://([^/]*)/?.*',
                               lambda m: m.groups()[0],
                               pulp.registry)

        # Store the registry URI in the push configuration
        self.workflow.push_conf.add_pulp_registry(self.pulp_registry_name,
                                                  pulp_registry)

        self.log.info("syncing from docker V2 registry %s",
                      self.docker_registry)

        images = []
        repos = {}  # pulp repo -> repo id
        with dockpulp_config(docker_registry=self.docker_registry) as config:
            for image in self.workflow.tag_conf.primary_images:
                if image.pulp_repo not in repos:
                    self.log.info("syncing %s", image.pulp_repo)
                    repoinfo = pulp.syncRepo(config.env, image.pulp_repo,
                                             config_file=config.name)
                    repos[image.pulp_repo] = repoinfo[0]['id']


                images.append(ImageName(registry=pulp_registry,
                                        repo=image.repo))

        self.log.info("publishing to crane")
        pulp.crane(list(repos.values()), wait=True)

        # Return the set of qualitifed repo names for this image
        return images



"""
Copyright (c) 2015 Red Hat, Inc
All rights reserved.

This software may be modified and distributed under the terms
of the BSD license. See the LICENSE file for details.


To have everything for a build in dist-git you need to fetch artefacts using 'fedpkg sources'.

This plugin should do it.
"""
import os
import subprocess

from atomic_reactor.plugin import PreBuildPlugin


class DistgitFetchArtefactsPlugin(PreBuildPlugin):
    key = "distgit_fetch_artefacts"
    is_allowed_to_fail = False

    def __init__(self, tasker, workflow, command):
        """
        constructor

        :param tasker: DockerTasker instance
        :param workflow: DockerBuildWorkflow instance
        :param command: str, command to use to get artefacts (e.g. 'make sources')
                             it is executed in cloned git repo
        """
        # call parent constructor
        super(DistgitFetchArtefactsPlugin, self).__init__(tasker, workflow)
        self.command = command

    def run(self):
        """
        fetch artefacts
        """
        source_path = self.workflow.source.path
        sources_file_path = os.path.join(source_path, 'sources')
        artefacts = ""
        try:
            with open(sources_file_path, 'r') as f:
                artefacts = f.read()
                self.log.info('sources file:\n%s', artefacts)
        except IOError as ex:
            if ex.errno == 2:
                self.log.info("no sources file")
            else:
                raise
        else:
            cur_dir = os.getcwd()
            os.chdir(source_path)
            subprocess.check_call(self.command.split())
            os.chdir(cur_dir)
        return artefacts



# -*- coding: utf-8 -*-
"""
Copyright (c) 2015 Red Hat, Inc
All rights reserved.

This software may be modified and distributed under the terms
of the BSD license. See the LICENSE file for details.
"""

from __future__ import unicode_literals

import os

HERE = os.path.dirname(__file__)
FILES = os.path.join(HERE, 'files')

MOCK = os.environ.get('NOMOCK') is None

INPUT_IMAGE = "busybox:latest"
DOCKERFILE_FILENAME = 'Dockerfile'
DOCKERFILE_GIT = "https://github.com/TomasTomecek/docker-hello-world.git"
DOCKERFILE_SHA1 = "6e592f1420efcd331cd28b360a7e02f669caf540"
DOCKERFILE_OK_PATH = os.path.join(FILES, 'docker-hello-world')
DOCKERFILE_ERROR_BUILD_PATH = os.path.join(FILES, 'docker-hello-world-error-build')
DOCKERFILE_SUBDIR_PATH = os.path.join(FILES, 'df-in-subdir')
DOCKERFILE_SHA1 = "6e592f1420efcd331cd28b360a7e02f669caf540"

SOURCE = {
    'provider': 'git',
    'uri': DOCKERFILE_GIT,
    'provider_params': {
        'git_commit': 'master',
    }
}

MOCK_SOURCE = {'provider': 'git', 'uri': 'asd'}

REGISTRY_PORT = "5000"
DOCKER0_IP = "172.17.42.1"
TEST_IMAGE = "atomic-reactor-test-image"

LOCALHOST_REGISTRY = "localhost:%s" % REGISTRY_PORT
DOCKER0_REGISTRY = "%s:%s" % (DOCKER0_IP, REGISTRY_PORT)
LOCALHOST_REGISTRY_HTTP = "http://%s" % LOCALHOST_REGISTRY
DOCKER0_REGISTRY_HTTP = "http://%s" % DOCKER0_REGISTRY

COMMAND = "eporeporjgpeorjgpeorjgpeorjgpeorjgpeorjg"



"""
Copyright (c) 2015 Red Hat, Inc
All rights reserved.

This software may be modified and distributed under the terms
of the BSD license. See the LICENSE file for details.


Python API for atomic_reactor. This is the official way of interacting with atomic_reactor.
"""
from atomic_reactor.inner import DockerBuildWorkflow
from atomic_reactor.outer import PrivilegedBuildManager, DockerhostBuildManager
from atomic_reactor.plugins.pre_pull_base_image import PullBaseImagePlugin
from atomic_reactor.plugins.post_tag_and_push import TagAndPushPlugin


__all__ = (
    'build_image_in_privileged_container',
    'build_image_using_hosts_docker',
    'build_image_here',
)

def _prepare_build_json(image, source, parent_registry, target_registries,
                        parent_registry_insecure, target_registries_insecure,
                        dont_pull_base_image, **kwargs):

    target_registries = target_registries or []
    registries = dict([(registry, {"insecure": target_registries_insecure})
                       for registry in target_registries])

    build_json = {
        "image": image,
        "source": source,
        "postbuild_plugins": [{
            "name": TagAndPushPlugin.key,
            "args": {
                "registries": registries
            }
        }]
    }

    if not dont_pull_base_image:
        build_json["prebuild_plugins"] = [{
            "name": PullBaseImagePlugin.key,
            "args": {
                "parent_registry": parent_registry,
                "parent_registry_insecure": parent_registry_insecure,
            }
        }]

    build_json.update(kwargs)
    return build_json


def build_image_in_privileged_container(build_image, source, image,
        parent_registry=None, target_registries=None, push_buildroot_to=None,
        parent_registry_insecure=False, target_registries_insecure=False,
        dont_pull_base_image=False, **kwargs):
    """
    build image from provided dockerfile (specified by `source`) in privileged container by
    running another docker instance inside the container

    :param build_image: str, image where target image should be built
    :param source: dict, where/how to get source code to put in image
    :param image: str, tag for built image ([registry/]image_name[:tag])
    :param parent_registry: str, registry to pull base image from
    :param target_registries: list of str, list of registries to push image to (might change in future)
    :param push_buildroot_to: str, repository where buildroot should be pushed
    :param parent_registry_insecure: bool, allow connecting to parent registry over plain http
    :param target_registries_insecure: bool, allow connecting to target registries over plain http
    :param dont_pull_base_image: bool, don't pull or update base image specified in dockerfile

    :return: BuildResults
    """
    build_json = _prepare_build_json(image, source, parent_registry, target_registries,
                                     parent_registry_insecure, target_registries_insecure,
                                     dont_pull_base_image, **kwargs)
    m = PrivilegedBuildManager(build_image, build_json)
    build_response = m.build()
    if push_buildroot_to:
        m.commit_buildroot()
        m.push_buildroot(push_buildroot_to)
    return build_response


def build_image_using_hosts_docker(build_image, source, image,
        parent_registry=None, target_registries=None, push_buildroot_to=None,
        parent_registry_insecure=False, target_registries_insecure=False,
        dont_pull_base_image=False, **kwargs):
    """
    build image from provided dockerfile (specified by `source`) in privileged container
    using docker from host

    :param build_image: str, image where target image should be built
    :param source: dict, where/how to get source code to put in image
    :param image: str, tag for built image ([registry/]image_name[:tag])
    :param parent_registry: str, registry to pull base image from
    :param target_registries: list of str, list of registries to push image to (might change in future)
    :param push_buildroot_to: str, repository where buildroot should be pushed
    :param parent_registry_insecure: bool, allow connecting to parent registry over plain http
    :param target_registries_insecure: bool, allow connecting to target registries over plain http
    :param dont_pull_base_image: bool, don't pull or update base image specified in dockerfile

    :return: BuildResults
    """
    build_json = _prepare_build_json(image, source, parent_registry, target_registries,
                                     parent_registry_insecure, target_registries_insecure,
                                     dont_pull_base_image, **kwargs)
    m = DockerhostBuildManager(build_image, build_json)
    build_response = m.build()
    if push_buildroot_to:
        m.commit_buildroot()
        m.push_buildroot(push_buildroot_to)
    return build_response


def build_image_here(source, image,
        parent_registry=None, target_registries=None, parent_registry_insecure=False,
        target_registries_insecure=False, dont_pull_base_image=False, **kwargs):
    """
    build image from provided dockerfile (specified by `source`) in current environment

    :param source: dict, where/how to get source code to put in image
    :param image: str, tag for built image ([registry/]image_name[:tag])
    :param parent_registry: str, registry to pull base image from
    :param target_registries: list of str, list of registries to push image to (might change in future)
    :param parent_registry_insecure: bool, allow connecting to parent registry over plain http
    :param target_registries_insecure: bool, allow connecting to target registries over plain http
    :param dont_pull_base_image: bool, don't pull or update base image specified in dockerfile

    :return: BuildResults
    """
    build_json = _prepare_build_json(image, source, parent_registry, target_registries,
                                     parent_registry_insecure, target_registries_insecure,
                                     dont_pull_base_image, **kwargs)
    m = DockerBuildWorkflow(**build_json)
    return m.build_docker_image()


def list_dockerfiles_in_git():
    """
    clone provided repo and return all dockerfiles found in the repo

    :return:
    """



"""
Copyright (c) 2015 Red Hat, Inc
All rights reserved.

This software may be modified and distributed under the terms
of the BSD license. See the LICENSE file for details.
"""

from __future__ import unicode_literals

from dockerfile_parse import DockerfileParser
from atomic_reactor.inner import DockerBuildWorkflow
from atomic_reactor.plugin import PreBuildPluginsRunner
from atomic_reactor.plugins.pre_return_dockerfile import CpDockerfilePlugin
from atomic_reactor.util import ImageName
from tests.constants import MOCK_SOURCE
from tests.fixtures import docker_tasker


class Y(object):
    pass

class X(object):
    image_id = "xxx"
    source = Y()
    source.dockerfile_path = None
    source.path = None
    base_image = ImageName(repo="qwe", tag="asd")

def test_returndockerfile_plugin(tmpdir):
    df_content = """
FROM fedora
RUN yum install -y python-django
CMD blabla"""
    df = DockerfileParser(str(tmpdir))
    df.content = df_content

    workflow = DockerBuildWorkflow(MOCK_SOURCE, 'test-image')
    workflow.builder = X
    workflow.builder.df_path = df.dockerfile_path
    workflow.builder.df_dir = str(tmpdir)

    runner = PreBuildPluginsRunner(
        docker_tasker,
        workflow,
        [{
            'name': CpDockerfilePlugin.key
        }]
    )
    runner.run()
    assert CpDockerfilePlugin.key is not None

    assert workflow.prebuild_results.get(CpDockerfilePlugin.key, "") == df_content



"""
Copyright (c) 2015 Red Hat, Inc
All rights reserved.

This software may be modified and distributed under the terms
of the BSD license. See the LICENSE file for details.
"""

from __future__ import unicode_literals

import os
import tempfile
import pytest
import six

try:
    from collections import OrderedDict
except ImportError:
    # Python 2.6
    from ordereddict import OrderedDict
import docker
from atomic_reactor.util import ImageName, \
    wait_for_command, clone_git_repo, LazyGit, figure_out_dockerfile, render_yum_repo, \
    process_substitutions, get_checksums, print_version_of_tools, get_version_of_tools, \
    get_preferred_label_key
from tests.constants import DOCKERFILE_GIT, INPUT_IMAGE, MOCK, DOCKERFILE_SHA1

if MOCK:
    from tests.docker_mock import mock_docker

TEST_DATA = {
    "repository.com/image-name": ImageName(registry="repository.com", repo="image-name"),
    "repository.com/prefix/image-name:1": ImageName(registry="repository.com",
                                                    namespace="prefix",
                                                    repo="image-name", tag="1"),
    "repository.com/prefix/image-name": ImageName(registry="repository.com",
                                                  namespace="prefix",
                                                  repo="image-name"),
    "image-name": ImageName(repo="image-name"),
    "registry:5000/image-name:latest": ImageName(registry="registry:5000",
                                                 repo="image-name", tag="latest"),
    "registry:5000/image-name": ImageName(registry="registry:5000", repo="image-name"),
    "fedora:20": ImageName(repo="fedora", tag="20"),
    "prefix/image-name:1": ImageName(namespace="prefix", repo="image-name", tag="1"),
    }


def test_image_name_parse():
    for inp, parsed in TEST_DATA.items():
        assert ImageName.parse(inp) == parsed


def test_image_name_format():
    for expected, image_name in TEST_DATA.items():
        assert image_name.to_str() == expected


def test_image_name_comparison():
    # make sure that both "==" and "!=" are implemented right on both Python major releases
    i1 = ImageName(registry='foo.com', namespace='spam', repo='bar', tag='1')
    i2 = ImageName(registry='foo.com', namespace='spam', repo='bar', tag='1')
    assert i1 == i2
    assert not i1 != i2

    i2 = ImageName(registry='foo.com', namespace='spam', repo='bar', tag='2')
    assert not i1 == i2
    assert i1 != i2


def test_wait_for_command():
    if MOCK:
        mock_docker()

    d = docker.Client()
    logs_gen = d.pull(INPUT_IMAGE, stream=True)
    assert wait_for_command(logs_gen) is not None


def test_clone_git_repo(tmpdir):
    tmpdir_path = str(tmpdir.realpath())
    commit_id = clone_git_repo(DOCKERFILE_GIT, tmpdir_path)
    assert commit_id is not None
    assert len(commit_id) == 40  # current git hashes are this long
    assert os.path.isdir(os.path.join(tmpdir_path, '.git'))


def test_clone_git_repo_by_sha1(tmpdir):
    tmpdir_path = str(tmpdir.realpath())
    commit_id = clone_git_repo(DOCKERFILE_GIT, tmpdir_path, commit=DOCKERFILE_SHA1)
    assert commit_id is not None
    print(six.text_type(commit_id))
    print(commit_id)
    assert six.text_type(commit_id, encoding="ascii") == six.text_type(DOCKERFILE_SHA1)
    assert len(commit_id) == 40  # current git hashes are this long
    assert os.path.isdir(os.path.join(tmpdir_path, '.git'))


def test_figure_out_dockerfile(tmpdir):
    tmpdir_path = str(tmpdir.realpath())
    clone_git_repo(DOCKERFILE_GIT, tmpdir_path)
    path, dir = figure_out_dockerfile(tmpdir_path)
    assert os.path.isfile(path)
    assert os.path.isdir(dir)


def test_lazy_git():
    lazy_git = LazyGit(git_url=DOCKERFILE_GIT)
    with lazy_git:
        assert lazy_git.git_path is not None
        assert lazy_git.commit_id is not None
        assert len(lazy_git.commit_id) == 40  # current git hashes are this long


def test_lazy_git_with_tmpdir(tmpdir):
    t = str(tmpdir.realpath())
    lazy_git = LazyGit(git_url=DOCKERFILE_GIT, tmpdir=t)
    assert lazy_git._tmpdir == t
    assert lazy_git.git_path is not None
    assert lazy_git.commit_id is not None
    assert len(lazy_git.commit_id) == 40  # current git hashes are this long


def test_render_yum_repo_unicode():
    yum_repo = OrderedDict((
        ("name", "asd"),
        ("baseurl", "http://example.com/$basearch/test.repo"),
        ("enabled", "1"),
        ("gpgcheck", "0"),
    ))
    rendered_repo = render_yum_repo(yum_repo)
    assert rendered_repo == """\
[asd]
name=asd
baseurl=http://example.com/\$basearch/test.repo
enabled=1
gpgcheck=0
"""


@pytest.mark.parametrize('dct, subst, expected', [
    ({'foo': 'bar'}, ['foo=spam'], {'foo': 'spam'}),
    ({'foo': 'bar'}, ['baz=spam'], {'foo': 'bar', 'baz': 'spam'}),
    ({'foo': 'bar'}, ['foo.bar=spam'], {'foo': {'bar': 'spam'}}),
    ({'foo': 'bar'}, ['spam.spam=spam'], {'foo': 'bar', 'spam': {'spam': 'spam'}}),


    ({'x_plugins': [{'name': 'a', 'args': {'b': 'c'}}]}, {'x_plugins.a.b': 'd'},
        {'x_plugins': [{'name': 'a', 'args': {'b': 'd'}}]}),
    # substituting plugins doesn't add new params
    ({'x_plugins': [{'name': 'a', 'args': {'b': 'c'}}]}, {'x_plugins.a.c': 'd'},
        {'x_plugins': [{'name': 'a', 'args': {'b': 'c'}}]}),
    ({'x_plugins': [{'name': 'a', 'args': {'b': 'c'}}]}, {'x_plugins.X': 'd'},
        ValueError()),
])
def test_process_substitutions(dct, subst, expected):
    if isinstance(expected, Exception):
        with pytest.raises(type(expected)):
            process_substitutions(dct, subst)
    else:
        process_substitutions(dct, subst)
        assert dct == expected


@pytest.mark.parametrize('content, algorithms, expected', [
    (b'abc', ['md5', 'sha256'],
                 {'md5sum': '900150983cd24fb0d6963f7d28e17f72',
                  'sha256sum': 'ba7816bf8f01cfea414140de5dae2223b00361a396177a9cb410ff61f20015ad'}),
    (b'abc', ['md5'], {'md5sum': '900150983cd24fb0d6963f7d28e17f72'}),
    (b'abc', [], {})
])
def test_get_hexdigests(tmpdir, content, algorithms, expected):
    with tempfile.NamedTemporaryFile(dir=str(tmpdir)) as tmpfile:
        tmpfile.write(content)
        tmpfile.flush()

        checksums = get_checksums(tmpfile.name, algorithms)
        assert checksums == expected


def test_get_versions_of_tools():
    response = get_version_of_tools()
    assert isinstance(response, list)
    for t in response:
        assert t["name"]
        assert t["version"]


def test_print_versions_of_tools():
    print_version_of_tools()


@pytest.mark.parametrize('labels, name, expected', [
    ({'name': 'foo', 'Name': 'foo'}, 'name', 'name'),
    ({'name': 'foo', 'Name': 'foo'}, 'Name', 'name'),
    ({'name': 'foo'}, 'Name', 'name'),
    ({'Name': 'foo'}, 'name', 'Name'),
    ({}, 'Name', 'name'),
    ({}, 'foobar', 'foobar')
])
def test_preferred_labels(labels, name, expected):
    result = get_preferred_label_key(labels, name)
    assert result == expected



"""
Copyright (c) 2015 Red Hat, Inc
All rights reserved.

This software may be modified and distributed under the terms
of the BSD license. See the LICENSE file for details.
"""

from email.mime.text import MIMEText
import os
import smtplib
import socket
try:
    from urlparse import urljoin
except ImportError:
    from urllib.parse import urljoin

from dockerfile_parse import DockerfileParser
import requests

from atomic_reactor.plugin import ExitPlugin, PluginFailedException
from atomic_reactor.plugins.pre_check_and_set_rebuild import is_rebuild
from atomic_reactor.source import GitSource


class SendMailPlugin(ExitPlugin):
    """This plugins sends notifications about build results.

    Example configuration (see arguments for init for detailed explanation):
        "exit_plugins": [{
                "name": "sendmail",
                "args": {
                    "send_on": ["auto_canceled", "auto_fail"],
                    "url": "https://openshift-instance.com",
                    "pdc_url": "https://pdc-instance.com",
                    # pdc_secret_path is filled in automatically by osbs-client
                    "pdc_secret_path": "/path/to/file/with/pdc/token",
                    "smtp_uri": "smtp-server.com",
                    "from_address": "osbs@mycompany.com",
                    "error_addresses": ["admin@mycompany.com"],
                    # optional arguments follow
                    "submitter": "John Smith <jsmith@mycompany.com>",
                    "pdc_verify_cert": true,
                    "pdc_component_df_label": "BZComponent",
                    "pdc_contact_role": "Devel_Owner"
                }
        }]
    """
    key = "sendmail"

    # symbolic constants for states
    MANUAL_SUCCESS = 'manual_success'
    MANUAL_FAIL = 'manual_fail'
    AUTO_SUCCESS = 'auto_success'
    AUTO_FAIL = 'auto_fail'
    AUTO_CANCELED = 'auto_canceled'

    allowed_states = set([MANUAL_SUCCESS, MANUAL_FAIL, AUTO_SUCCESS, AUTO_FAIL, AUTO_CANCELED])

    PDC_TOKEN_FILE = 'pdc.token'
    PDC_CONTACT_ROLE = 'Devel_Owner'

    def __init__(self, tasker, workflow, send_on=None, url=None, submitter='unknown', pdc_url=None,
                 pdc_verify_cert=True, pdc_component_df_label="BZComponent", pdc_secret_path=None,
                 pdc_contact_role=None, smtp_uri=None, from_address=None,
                 error_addresses=None):
        """
        constructor

        :param tasker: DockerTasker instance
        :param workflow: DockerBuildWorkflow instance
        :param send_on: list of build states when a notification should be sent
        :param url: URL to OSv3 instance where the build logs are stored
        :param submitter: name of user who submitted a build (plain string)
        :param pdc_url: URL of PDC to query for contact information
        :param pdc_verify_cert: whether or not to verify SSL cert of PDC (defaults to True)
        :param pdc_component_df_label: name of Dockerfile label to use as PDC global_component
        :param pdc_secret_path: path to pdc.token file; $SOURCE_SECRET_PATH otherwise
        :param pdc_contact_role: name of PDC role to contact
        :param smtp_uri: URL of SMTP server to use to send the message (e.g. "foo.com:25")
        :param from_address: the "From" of the notification email
        :param error_addresses: list of email addresses where to send an email if there's an error
            (e.g. if we can't find out who to notify about the failed build)
        """
        super(SendMailPlugin, self).__init__(tasker, workflow)
        self.send_on = send_on
        self.url = url
        self.submitter = submitter
        self.pdc_url = pdc_url
        self.pdc_verify_cert = pdc_verify_cert
        self.pdc_component_df_label = pdc_component_df_label
        self.pdc_secret_path = pdc_secret_path
        self.pdc_contact_role = pdc_contact_role or self.PDC_CONTACT_ROLE
        self.smtp_uri = smtp_uri
        self.from_address = from_address
        self.error_addresses = error_addresses

    def _should_send(self, rebuild, success, canceled):
        """Return True if any state in `self.send_on` meets given conditions, thus meaning
        that a notification mail should be sent.
        """
        should_send = False

        should_send_mapping = {
            self.MANUAL_SUCCESS: not rebuild and success,
            self.MANUAL_FAIL: not rebuild and not success,
            self.AUTO_SUCCESS: rebuild and success,
            self.AUTO_FAIL: rebuild and not success,
            self.AUTO_CANCELED: rebuild and canceled
        }

        for state in self.send_on:
            should_send |= should_send_mapping[state]
        return should_send

    def _render_mail(self, rebuild, success, canceled):
        """Render and return subject and body of the mail to send."""
        subject_template = 'Image %(image)s; Status %(endstate)s; Submitted by %(user)s'
        body_template = '\n'.join([
            'Image: %(image)s',
            'Status: %(endstate)s',
            'Submitted by: %(user)s',
            'Logs: %(logs)s',
        ])

        endstate = None
        if canceled:
            endstate = 'canceled'
        else:
            endstate = 'successful' if success else 'failed'
        url = None
        if self.url and self.workflow.openshift_build_selflink:
            url = urljoin(self.url, self.workflow.openshift_build_selflink + '/log')

        formatting_dict = {
            'image': self.workflow.image,
            'endstate': endstate,
            'user': '<autorebuild>' if rebuild else self.submitter,
            'logs': url
        }
        return (subject_template % formatting_dict, body_template % formatting_dict)

    def _get_pdc_token(self):
        # we want to allow pdc_secret_path to be None in __init__ - I'm assuming that in future
        #  we'll want different sources of contact info, so we only want to raise when
        #  the plugin actually tries to authenticate against PDC and doesn't have pdc_secret_path
        if self.pdc_secret_path is None:
            raise PluginFailedException('Getting PDC token, but pdc_secret_path is unspecified')
        token_file = os.path.join(self.pdc_secret_path, self.PDC_TOKEN_FILE)

        self.log.debug('getting PDC token from file %s', token_file)

        with open(token_file, 'r') as f:
            return f.read().strip()

    def _get_component_label(self):
        """Get value of Dockerfile label that is to be used as `global_component` to query
        PDC release-components API endpoint.
        """
        labels = DockerfileParser(self.workflow.builder.df_path).labels
        if self.pdc_component_df_label not in labels:
            raise PluginFailedException('No %s label in Dockerfile, can\'t get PDC component',
                                        self.pdc_component_df_label)
        return labels[self.pdc_component_df_label]

    def _get_receivers_list(self):
        """Return list of receivers of the notification.

        :raises RuntimeError: if PDC can't be contacted or doesn't provide sufficient data
        :raises PluginFailedException: if there's a critical error while getting PDC data
        """

        # TODO: document what this plugin expects to be in Dockerfile/where it gets info from
        global_component = self._get_component_label()
        # this relies on bump_release plugin configuring source.git_commit to actually be
        #  branch name, not a commit
        if not isinstance(self.workflow.source, GitSource):
            raise PluginFailedException('Source is not of type "GitSource", panic!')
        git_branch = self.workflow.source.git_commit
        try:
            r = requests.get(urljoin(self.pdc_url, 'rest_api/v1/release-component-contacts/'),
                             headers={'Authorization': 'Token %s' % self._get_pdc_token()},
                             params={'global_component': global_component,
                                     'dist_git_branch': git_branch,
                                     'role': self.pdc_contact_role},
                             verify=self.pdc_verify_cert)
        except requests.RequestException as e:
            self.log.error('failed to connect to PDC: %s', str(e))
            raise RuntimeError(e)

        if r.status_code != 200:
            self.log.error('PDC returned status code %s, full response: %s',
                           r.status_code, r.text)
            raise RuntimeError('PDC returned non-200 status code (%s), see referenced build log' %
                               r.status_code)

        contacts = r.json()

        if contacts['count'] == 0:
            self.log.error('no %s role for the component', self.pdc_contact_role)
            raise RuntimeError('no %s role for the component' % self.pdc_contact_role)

        send_to = []
        for contact in contacts['results']:
            send_to.append(contact['contact']['email'])

        return send_to

    def _send_mail(self, receivers_list, subject, body):
        """Actually sends the mail with `subject` and `body` to all members of `receivers_list`."""
        msg = MIMEText(body)
        msg['Subject'] = subject
        msg['From'] = self.from_address
        msg['To'] = ', '.join(receivers_list)

        s = None
        try:
            s = smtplib.SMTP(self.smtp_uri)
            s.sendmail(self.from_address, receivers_list, msg.as_string())
        except (socket.gaierror, smtplib.SMTPException) as e:
            raise PluginFailedException('Error communicating with SMTP server: %s' % str(e))
        finally:
            if s is not None:
                s.quit()

    def run(self):
        # verify that given states are subset of allowed states
        unknown_states = set(self.send_on) - self.allowed_states
        if len(unknown_states) > 0:
            raise PluginFailedException('Unknown state(s) "%s" for sendmail plugin' %
                                        '", "'.join(sorted(unknown_states)))

        rebuild = is_rebuild(self.workflow)
        success = not self.workflow.build_failed
        canceled = self.workflow.autorebuild_canceled

        self.log.info('checking conditions for sending notification ...')
        if self._should_send(rebuild, success, canceled):
            self.log.info('notification about build result will be sent')
            subject, body = self._render_mail(rebuild, success, canceled)
            try:
                self.log.debug('getting list of receivers for this component ...')
                receivers = self._get_receivers_list()
            except RuntimeError as e:
                self.log.error('couldn\'t get list of receivers, sending error message ...')
                # TODO: maybe improve the error message/subject
                body = '\n'.join([
                    'Failed to get contact for %s, error: %s' % (str(self.workflow.image), str(e)),
                    'Since your address is in "error_addresses", this email was sent to you to '
                    'take action on this.',
                    'Wanted to send following mail:',
                    '',
                    body
                ])
                receivers = self.error_addresses
            self.log.info('sending notification to %s ...', receivers)
            self._send_mail(receivers, subject, body)
        else:
            self.log.info('conditions for sending notification not met, doing nothing')



"""
Copyright (c) 2015 Red Hat, Inc
All rights reserved.

This software may be modified and distributed under the terms
of the BSD license. See the LICENSE file for details.


Reads input from OpenShift v3
"""
import json
import os

from atomic_reactor.plugin import InputPlugin
from atomic_reactor.plugins.pre_pull_base_image import PullBaseImagePlugin
from atomic_reactor.plugins.post_tag_and_push import TagAndPushPlugin


class OSv3InputPlugin(InputPlugin):
    key = "osv3"

    def __init__(self, **kwargs):
        """
        constructor
        """
        # call parent constructor
        super(OSv3InputPlugin, self).__init__(**kwargs)

    def run(self):
        """
        each plugin has to implement this method -- it is used to run the plugin actually

        response from plugin is kept and used in json result response
        """
        build_json_str = os.environ['BUILD']
        build_json = json.loads(build_json_str)
        git_url = os.environ['SOURCE_URI']
        git_ref = os.environ.get('SOURCE_REF', None)
        image = os.environ['OUTPUT_IMAGE']
        self.target_registry = os.environ.get('OUTPUT_REGISTRY', None)

        try:
            self.plugins_json = os.environ['ATOMIC_REACTOR_PLUGINS']
        except KeyError:
            try:
                self.plugins_json = os.environ['DOCK_PLUGINS']
            except KeyError:
                raise RuntimeError("No plugin configuration found!")
            else:
                self.log.warning("DOCK_PLUGINS is deprecated - please update your osbs-client!")

        self.plugins_json = json.loads(self.plugins_json)

        self.preprocess_plugins()

        input_json = {
            'source': {
                'provider': 'git',
                'uri': git_url,
                'provider_params': {'git_commit': git_ref}
            },
            'image': image,
            'openshift_build_selflink': build_json.get('metadata', {}).get('selfLink', None)
        }
        input_json.update(self.plugins_json)

        self.log.debug("build json: %s", input_json)

        return input_json

    def get_plugin(self, section, name):
        try:
            plugin = [p for p in self.plugins_json[section] if p.get('name') == name][0]
        except IndexError:
            return None
        else:
            return plugin

    def preprocess_plugins(self):
        for typ in ('prebuild', 'postbuild', 'prepublish', 'exit'):
            self.plugins_json.setdefault('{0}_plugins'.format(typ), [])

        # XXX: Remove the two if blocks after we stop using super old osbs:(
        pull_plugin = self.get_plugin('prebuild_plugins', PullBaseImagePlugin.key)
        if not pull_plugin:
            self.log.warning("%s is missing in prebuild_plugins - please update your osbs-client!",
                             PullBaseImagePlugin.key)
            pull_plugin = { "name": PullBaseImagePlugin.key }
            self.plugins_json['prebuild_plugins'].insert(0, pull_plugin)

        change_plugin = self.get_plugin('prebuild_plugins', 'change_source_registry')
        if change_plugin:
            if 'registry_uri' in change_plugin.get('args', {}):
                pull_plugin.setdefault('args', {})['parent_registry'] = \
                    change_plugin['args']['registry_uri']
            if 'insecure_registry' in change_plugin.get('args', {}):
                pull_plugin.setdefault('args', {})['parent_registry_insecure'] = \
                    change_plugin['args']['insecure_registry']
            self.plugins_json['prebuild_plugins'].remove(change_plugin)

        if 'parent_registry' not in pull_plugin.get('args', {}):
            self.log.error("source registry is not configured")

        push_plugin = self.get_plugin('postbuild_plugins', TagAndPushPlugin.key)
        if not push_plugin and self.target_registry:
            self.log.warning("%s is missing in postbuild_plugins - please update your osbs-client!",
                             TagAndPushPlugin.key)
            push_plugin = { "name": TagAndPushPlugin.key }
            self.plugins_json['postbuild_plugins'].insert(0, push_plugin)

        if push_plugin and self.target_registry:
            push_plugin.setdefault('args', {}).setdefault('registries', {})
            if not push_plugin['args']['registries']:
                push_plugin['args']['registries'][self.target_registry] = {"insecure": True}

    @classmethod
    def is_autousable(cls):
        return 'BUILD' in os.environ and 'SOURCE_URI' in os.environ and 'OUTPUT_IMAGE' in os.environ



# -*- coding: utf-8 -*-
"""
Copyright (c) 2015 Red Hat, Inc
All rights reserved.

This software may be modified and distributed under the terms
of the BSD license. See the LICENSE file for details.
"""

from __future__ import unicode_literals

from six import string_types


def is_string_type(obj):
    """
    Test whether obj is a string type

    :param obj: object to test
    :return: bool, whether obj is a string type
    """

    return any(isinstance(obj, strtype)
               for strtype in string_types)



"""
Copyright (c) 2015 Red Hat, Inc
All rights reserved.

This software may be modified and distributed under the terms
of the BSD license. See the LICENSE file for details.
"""

from __future__ import unicode_literals

import docker
from flexmock import flexmock
import pytest

from atomic_reactor.core import DockerTasker
from atomic_reactor.inner import DockerBuildWorkflow
from atomic_reactor.plugin import PostBuildPluginsRunner
from atomic_reactor.plugins.post_rpmqa import PostBuildRPMqaPlugin
from atomic_reactor.util import ImageName
from tests.constants import DOCKERFILE_GIT, MOCK
if MOCK:
    from tests.docker_mock import mock_docker
from tests.fixtures import docker_tasker

TEST_IMAGE = "fedora:latest"
SOURCE = {"provider": "git", "uri": DOCKERFILE_GIT}


class X(object):
    pass


PACKAGE_LIST = ['python-docker-py,1.3.1,1.fc24,noarch,(none),191456,7c1f60d8cde73e97a45e0c489f4a3b26,1438058212',
                'fedora-repos-rawhide,24,0.1,noarch,(none),2149,d41df1e059544d906363605d47477e60,1436940126',
                'gpg-pubkey-doc,1.0,1,noarch,(none),1000,00000000000000000000000000000000,1436940126']
PACKAGE_LIST_WITH_AUTOGENERATED = PACKAGE_LIST + ['gpg-pubkey,qwe123,zxcasd123,(none),(none),0,(none),1370645731']
PACKAGE_LIST_WITH_AUTOGENERATED_B = [x.encode("utf-8") for x in PACKAGE_LIST_WITH_AUTOGENERATED]


def mock_logs(cid, **kwargs):
    return b"\n".join(PACKAGE_LIST_WITH_AUTOGENERATED_B)


@pytest.mark.parametrize("ignore_autogenerated", [
    {"ignore": True, "package_list": PACKAGE_LIST},
    {"ignore": False, "package_list": PACKAGE_LIST_WITH_AUTOGENERATED},
])
def test_rpmqa_plugin(ignore_autogenerated):
    if MOCK:
        mock_docker()
    tasker = DockerTasker()
    workflow = DockerBuildWorkflow(SOURCE, "test-image")
    setattr(workflow, 'builder', X())
    setattr(workflow.builder, 'image_id', "asd123")
    setattr(workflow.builder, 'base_image', ImageName(repo='fedora', tag='21'))
    setattr(workflow.builder, "source", X())
    setattr(workflow.builder.source, 'dockerfile_path', "/non/existent")
    setattr(workflow.builder.source, 'path', "/non/existent")

    flexmock(docker.Client, logs=mock_logs)
    runner = PostBuildPluginsRunner(
        tasker,
        workflow,
        [{"name": PostBuildRPMqaPlugin.key,
          "args": {
              'image_id': TEST_IMAGE,
              "ignore_autogenerated_gpg_keys": ignore_autogenerated["ignore"]}}
    ])
    results = runner.run()
    assert results[PostBuildRPMqaPlugin.key] == ignore_autogenerated["package_list"]


def mock_logs_raise(cid, **kwargs):
    raise RuntimeError


def test_rpmqa_plugin_exception(docker_tasker):
    workflow = DockerBuildWorkflow(SOURCE, "test-image")
    setattr(workflow, 'builder', X())
    setattr(workflow.builder, 'image_id', "asd123")
    setattr(workflow.builder, 'base_image', ImageName(repo='fedora', tag='21'))
    setattr(workflow.builder, "source", X())
    setattr(workflow.builder.source, 'dockerfile_path', "/non/existent")
    setattr(workflow.builder.source, 'path', "/non/existent")

    flexmock(docker.Client, logs=mock_logs_raise)
    runner = PostBuildPluginsRunner(docker_tasker, workflow,
                                    [{"name": PostBuildRPMqaPlugin.key,
                                      "args": {'image_id': TEST_IMAGE}}])
    results = runner.run()

    assert results is not None
    assert results[PostBuildRPMqaPlugin.key] is not None
    assert isinstance(results[PostBuildRPMqaPlugin.key], Exception)



"""
Copyright (c) 2015 Red Hat, Inc
All rights reserved.

This software may be modified and distributed under the terms
of the BSD license. See the LICENSE file for details.
"""

import json
from atomic_reactor.constants import CONTAINER_RESULTS_JSON_PATH
from atomic_reactor.inner import BuildResultsEncoder
from atomic_reactor.plugin import ExitPlugin


__all__ = ('StoreLogsToFilePlugin', )


class StoreLogsToFilePlugin(ExitPlugin):
    key = "store_logs_to_file"

    def __init__(self, tasker, workflow, file_path):
        """
        constructor

        :param tasker: DockerTasker instance
        :param workflow: DockerBuildWorkflow instance
        :param file_path: str, path to file where logs should be stored
        """
        # call parent constructor
        super(StoreLogsToFilePlugin, self).__init__(tasker, workflow)
        self.file_path = file_path

    def run(self):
        file_path = self.file_path or CONTAINER_RESULTS_JSON_PATH
        results = {
            'prebuild_plugins': self.workflow.prebuild_results,
            'postbuild_plugins': self.workflow.postbuild_results,
        }

        with open(file_path, 'w') as results_json_fd:
            json.dump(results, results_json_fd, cls=BuildResultsEncoder)



#!/usr/bin/python
"""
Copyright (c) 2015 Red Hat, Inc
All rights reserved.

This software may be modified and distributed under the terms
of the BSD license. See the LICENSE file for details.
"""

import re
import sys

from setuptools import setup, find_packages
from atomic_reactor.constants import DESCRIPTION, HOMEPAGE

data_files = {
    "share/atomic-reactor/images/privileged-builder": [
        "images/privileged-builder/Dockerfile",
        "images/privileged-builder/docker.sh",
    ],
    "share/atomic-reactor/images/dockerhost-builder": [
        "images/dockerhost-builder/Dockerfile",
    ],
}

def _get_requirements(path):
    try:
        with open(path) as f:
            packages = f.read().splitlines()
    except (IOError, OSError) as ex:
        raise RuntimeError("Can't open file with requirements: %r", ex)
    packages = (p.strip() for p in packages if not re.match("^\s*#", p))
    packages = list(filter(None, packages))
    return packages

def _install_requirements():
    requirements = _get_requirements('requirements.txt')
    if sys.version_info[0] < 3:
        requirements += _get_requirements('requirements-py2.txt')
    return requirements

setup(
    name='atomic-reactor',
    version='1.6.1',
    description=DESCRIPTION,
    author='Red Hat, Inc.',
    author_email='atomic-devel@projectatomic.io',
    url=HOMEPAGE,
    license="BSD",
    entry_points={
        'console_scripts': ['atomic-reactor=atomic_reactor.cli.main:run',
                            'pulpsecret-gen=atomic_reactor.cli.secret:run'],
    },
    packages=find_packages(exclude=["*.tests", "*.tests.*", "tests.*", "tests"]),
    install_requires=_install_requirements(),
    data_files=data_files.items(),
)




"""
Copyright (c) 2015 Red Hat, Inc
All rights reserved.

This software may be modified and distributed under the terms
of the BSD license. See the LICENSE file for details.
"""

from __future__ import unicode_literals

import pytest
from atomic_reactor.core import DockerTasker
from atomic_reactor.inner import DockerBuildWorkflow
from atomic_reactor.plugin import PreBuildPluginsRunner, PluginFailedException
from atomic_reactor.plugins.pre_check_and_set_rebuild import (is_rebuild,
                                                              CheckAndSetRebuildPlugin)
from atomic_reactor.util import ImageName
import json
from osbs.api import OSBS
from osbs.exceptions import OsbsResponseException
from flexmock import flexmock
from tests.constants import SOURCE, MOCK
if MOCK:
    from tests.docker_mock import mock_docker


class X(object):
    pass


class TestCheckRebuild(object):
    def prepare(self, key, value, set_labels_args=None, set_labels_kwargs=None):
        if MOCK:
            mock_docker()
        tasker = DockerTasker()
        workflow = DockerBuildWorkflow(SOURCE, "test-image")
        setattr(workflow, 'builder', X())
        setattr(workflow.builder, 'image_id', 'asd123')
        setattr(workflow.builder, 'base_image', ImageName(repo='Fedora',
                                                          tag='22'))
        setattr(workflow.builder, 'source', X())
        setattr(workflow.builder.source, 'path', '/tmp')
        setattr(workflow.builder.source, 'dockerfile_path', None)
        expectation = (flexmock(OSBS)
                       .should_receive('set_labels_on_build_config'))
        if set_labels_args is not None:
            if set_labels_kwargs is None:
                set_labels_kwargs = {}

            expectation.with_args(*set_labels_args, **set_labels_kwargs)

        runner = PreBuildPluginsRunner(tasker, workflow, [
            {
                'name': CheckAndSetRebuildPlugin.key,
                'args': {
                    'label_key': key,
                    'label_value': value,
                    'url': '',
                },
            }
        ])
        return workflow, runner

    def test_check_rebuild_no_build_json(self, monkeypatch):
        workflow, runner = self.prepare('is_autorebuild', 'true')
        monkeypatch.delenv('BUILD', raising=False)

        with pytest.raises(PluginFailedException):
            runner.run()

    def test_check_no_buildconfig(self, monkeypatch):
        key = 'is_autorebuild'
        value = 'true'
        workflow, runner = self.prepare(key, value)
        monkeypatch.setenv("BUILD", json.dumps({
            "metadata": {
                "labels": {
                    key: value,
                }
            }
        }))

        # No buildconfig in metadata
        with pytest.raises(PluginFailedException):
            runner.run()

    @pytest.mark.parametrize(('namespace'), [None, 'my_namespace'])
    def test_check_is_not_rebuild(self, namespace, monkeypatch):
        key = 'is_autorebuild'
        value = 'true'
        buildconfig = "buildconfig1"
        namespace_dict = {}
        if namespace is not None:
            namespace_dict["namespace"] = namespace

        workflow, runner = self.prepare(key, value,
                                        set_labels_args=(buildconfig,
                                                         {key: value}),
                                        set_labels_kwargs=namespace_dict)

        build_json = {
            "metadata": {
                "labels": {
                    "buildconfig": buildconfig,
                    key: "false",
                }
            }
        }
 
        build_json["metadata"].update(namespace_dict)
        monkeypatch.setenv("BUILD", json.dumps(build_json))
        runner.run()
        assert workflow.prebuild_results[CheckAndSetRebuildPlugin.key] == False
        assert not is_rebuild(workflow)

    def test_check_is_rebuild(self, monkeypatch):
        key = 'is_autorebuild'
        value = 'true'
        workflow, runner = self.prepare(key, value)

        monkeypatch.setenv("BUILD", json.dumps({
            "metadata": {
                "labels": {
                    "buildconfig": "buildconfig1",
                    key: value,
                }
            }
        }))
        runner.run()
        assert workflow.prebuild_results[CheckAndSetRebuildPlugin.key] == True
        assert is_rebuild(workflow)

    def test_409_response(self, monkeypatch):
        key = 'is_autorebuild'
        workflow, runner = self.prepare(key, 'true')

        (flexmock(OSBS)
            .should_receive('set_labels_on_build_config')
            .twice()
            .and_raise(OsbsResponseException('conflict', 409))
            .and_return(None))

        monkeypatch.setenv("BUILD", json.dumps({
            "metadata": {
                "labels": {
                    "buildconfig": "buildconfig1",
                    key: 'false',
                }
            }
        }))
        runner.run()



"""
Copyright (c) 2015 Red Hat, Inc
All rights reserved.

This software may be modified and distributed under the terms
of the BSD license. See the LICENSE file for details.


Naming Conventions
==================

registry.somewhere/namespace/image_name:tag
|-----------------|                          registry, reg_uri
                  |---------|                namespace
|--------------------------------------|     repository
                  |--------------------|     image name
                                        |--| tag
                  |------------------------| image
|------------------------------------------| image

I've tried to be as much consistent (man pages were source) with docker as possible


"""
import os
import shutil
import logging
import tempfile
import json

import docker
from docker.errors import APIError

from atomic_reactor.constants import CONTAINER_SHARE_PATH, CONTAINER_SHARE_SOURCE_SUBDIR,\
        BUILD_JSON, DOCKER_SOCKET_PATH
from atomic_reactor.source import get_source_instance_for
from atomic_reactor.util import ImageName, wait_for_command, clone_git_repo, figure_out_dockerfile


logger = logging.getLogger(__name__)


class LastLogger(object):
    """
    provide method for getting last log
    """

    def __init__(self):
        self._last_logs = []

    @property
    def last_logs(self):
        """ logs from last operation """
        return self._last_logs

    @last_logs.setter
    def last_logs(self, value):
        self._last_logs = value


class BuildContainerFactory(object):
    """
    set of methods for building images inside containers
    """

    def __init__(self):
        self.tasker = DockerTasker()

    def _check_build_input(self, image, args_path):
        """
        Internal method, validate provided args.

        :param image: str
        :param args_path: str, path dir which is mounter inside container
        :return: None
        :raises RuntimeError
        """
        try:
            with open(os.path.join(args_path, BUILD_JSON)) as json_args:
                logger.debug("build input: image = '%s', args = '%s'", image, json_args.read())
        except (IOError, OSError) as ex:
            logger.error("unable to open json arguments: %r", ex)
            raise RuntimeError("Unable to open json arguments: %r" % ex)

        if not self.tasker.image_exists(image):
            logger.error("provided build image doesn't exist: '%s'", image)
            raise RuntimeError("Provided build image doesn't exist: '%s'" % image)

    def _obtain_source_from_path_if_needed(self, local_path, container_path=CONTAINER_SHARE_PATH):
        # TODO: maybe we should do this for any provider? (if we expand to various providers
        #  like mercurial, we don't to force container to have mercurial installed, etc.)
        build_json_path = os.path.join(local_path, BUILD_JSON)
        with open(build_json_path, 'r') as fp:
            build_json = json.load(fp)
        source = get_source_instance_for(build_json['source'], tmpdir=local_path)
        if source.provider == 'path':
            logger.debug('copying source from %s to %s', source.schemeless_path, local_path)
            source.get()
            logger.debug('verifying that %s exists: %s', local_path, os.path.exists(local_path))
            # now modify the build json
            build_json['source']['uri'] =\
                    'file://' + os.path.join(container_path, CONTAINER_SHARE_SOURCE_SUBDIR)
            with open(build_json_path, 'w') as fp:
                json.dump(build_json, fp)
        # else we do nothing

    @staticmethod
    def _volume_bind_understands_mode():
        # docker.utils.convert_volume_binds() understands 'mode' since docker-py-1.3.0
        # returns ['/a/:/b/:rw'] with docker-py < 1.3.0 and ['/a/:/b/:ro,Z'] with >= 1.3.0
        bind = docker.utils.convert_volume_binds({'/a/': {'bind': '/b/', 'mode': 'ro,Z'}})
        if bind and len(bind[0].split(':')) > 2 and bind[0].split(':')[2] == 'ro,Z':
            return True
        return False

    def build_image_dockerhost(self, build_image, json_args_path):
        """
        Build docker image inside privileged container using docker from host
        (mount docker socket inside container).
        There are possible races here. Use wisely.

        This operation is asynchronous and you should wait for container to finish.

        :param build_image: str, name of image where build is performed
        :param json_args_path: str, this dir is mounted inside build container and used
                               as a way to transport data between host and buildroot; there
                               has to be a file inside this dir with name atomic_reactor.BUILD_JSON which
                               is used to feed build
        :return: str, container id
        """
        logger.info("building image '%s' in container using docker from host", build_image)

        self._check_build_input(build_image, json_args_path)
        self._obtain_source_from_path_if_needed(json_args_path, CONTAINER_SHARE_PATH)

        if not os.path.exists(DOCKER_SOCKET_PATH):
            logger.error("looks like docker is not running because there is no socket at: %s", DOCKER_SOCKET_PATH)
            raise RuntimeError("docker socket not found: %s" % DOCKER_SOCKET_PATH)

        volume_bindings = {
            DOCKER_SOCKET_PATH: {
                'bind': DOCKER_SOCKET_PATH,
            },
            json_args_path: {
                'bind': CONTAINER_SHARE_PATH,
            },
        }

        if self._volume_bind_understands_mode():
            volume_bindings[DOCKER_SOCKET_PATH]['mode'] = 'ro'
            volume_bindings[json_args_path]['mode'] = 'rw,Z'
        else:
            volume_bindings[DOCKER_SOCKET_PATH]['ro'] = True
            volume_bindings[json_args_path]['rw'] = True

        logger.debug('build json mounted in container: %s',
                open(os.path.join(json_args_path, BUILD_JSON)).read())
        container_id = self.tasker.run(
            ImageName.parse(build_image),
            create_kwargs={'volumes': [DOCKER_SOCKET_PATH, json_args_path]},
            start_kwargs={'binds': volume_bindings,
                          'privileged': True}
        )

        return container_id

    def build_image_privileged_container(self, build_image, json_args_path):
        """
        Build image inside privileged container: this will run another docker instance inside

        This operation is asynchronous and you should wait for container to finish.

        :param build_image: str, name of image where build is performed
        :param json_args_path: str, this dir is mounted inside build container and used
                               as a way to transport data between host and buildroot; there
                               has to be a file inside this dir with name atomic_reactor.BUILD_JSON which
                               is used to feed build
        :return: dict, keys container_id and stream
        """
        logger.info("building image '%s' inside privileged container", build_image)

        self._check_build_input(build_image, json_args_path)
        self._obtain_source_from_path_if_needed(json_args_path, CONTAINER_SHARE_PATH)

        volume_bindings = {
            json_args_path: {
                'bind': CONTAINER_SHARE_PATH,
            },
        }

        if self._volume_bind_understands_mode():
            volume_bindings[json_args_path]['mode'] = 'rw,Z'
        else:
            volume_bindings[json_args_path]['rw'] = True

        logger.debug('build json mounted in container: %s',
                open(os.path.join(json_args_path, BUILD_JSON)).read())
        container_id = self.tasker.run(
            ImageName.parse(build_image),
            create_kwargs={'volumes': [json_args_path]},
            start_kwargs={'binds': volume_bindings,
                          'privileged': True}
        )

        return container_id


class DockerTasker(LastLogger):
    def __init__(self, base_url=None, timeout=120, **kwargs):
        """
        Constructor

        :param base_url: str, docker connection URL
        :param timeout: int, timeout for docker client
        """
        super(DockerTasker, self).__init__(**kwargs)

        client_kwargs = {}
        if base_url:
            client_kwargs['base_url'] = base_url
        elif os.environ.get('DOCKER_CONNECTION'):
            client_kwargs['base_url'] = os.environ['DOCKER_CONNECTION']

        if hasattr(docker, 'AutoVersionClient'):
            client_kwargs['version'] = 'auto'

        self.d = docker.Client(timeout=timeout, **client_kwargs)

    def build_image_from_path(self, path, image, stream=False, use_cache=False, remove_im=True):
        """
        build image from provided path and tag it

        this operation is asynchronous and you should consume returned generator in order to wait
        for build to finish

        :param path: str
        :param image: ImageName, name of the resulting image
        :param stream: bool, True returns generator, False returns str
        :param use_cache: bool, True if you want to use cache
        :param remove_im: bool, remove intermediate containers produced during docker build
        :return: generator
        """
        logger.info("building image '%s' from path '%s'", image, path)
        try:
            response = self.d.build(path=path, tag=image.to_str(), stream=stream, nocache=not use_cache,
                                    rm=remove_im, pull=False)  # returns generator
        except TypeError:
            # because changing api is fun
            response = self.d.build(path=path, tag=image.to_str(), stream=stream, nocache=not use_cache,
                                    rm=remove_im)  # returns generator
        return response

    def build_image_from_git(self, url, image, git_path=None, git_commit=None, copy_dockerfile_to=None,
                             stream=False, use_cache=False):
        """
        build image from provided url and tag it

        this operation is asynchronous and you should consume returned generator in order to wait
        for build to finish

        :param url: str
        :param image: ImageName, name of the resulting image
        :param git_path: str, path to dockerfile within gitrepo
        :param copy_dockerfile_to: str, copy dockerfile to provided path
        :param stream: bool, True returns generator, False returns str
        :param use_cache: bool, True if you want to use cache
        :return: generator
        """
        logger.info("building image '%s' from git repo '%s' specified as URL '%s'",
                image, git_path, url)
        logger.info("will copy Dockerfile to '%s'", copy_dockerfile_to)
        temp_dir = tempfile.mkdtemp()
        response = None
        try:
            clone_git_repo(url, temp_dir, git_commit)
            df_path, df_dir = figure_out_dockerfile(temp_dir, git_path)
            if copy_dockerfile_to:  # TODO: pre build plugin
                shutil.copyfile(df_path, copy_dockerfile_to)
            response = self.build_image_from_path(df_dir, image, stream=stream, use_cache=use_cache)
        finally:
            try:
                shutil.rmtree(temp_dir)
            except (IOError, OSError) as ex:
                # no idea why this is happening
                logger.warning("Failed to remove dir '%s': %r", temp_dir, ex)
        logger.info("build finished")
        return response

    def run(self, image, command=None, create_kwargs=None, start_kwargs=None):
        """
        create container from provided image and start it

        for more info, see documentation of REST API calls:
         * containers/{}/start
         * container/create

        :param image: ImageName or string, name or id of the image
        :param command: str
        :param create_kwargs: dict, kwargs for docker.create_container
        :param start_kwargs: dict, kwargs for docker.start
        :return: str, container id
        """
        logger.info("creating container from image '%s' and running it", image)
        create_kwargs = create_kwargs or {}
        start_kwargs = start_kwargs or {}
        logger.debug("image = '%s', command = '%s', create_kwargs = '%s', start_kwargs = '%s'",
                     image, command, create_kwargs, start_kwargs)
        if isinstance(image, ImageName):
            image = image.to_str()
        container_dict = self.d.create_container(image, command=command, **create_kwargs)
        container_id = container_dict['Id']
        logger.debug("container_id = '%s'", container_id)
        self.d.start(container_id, **start_kwargs)  # returns None
        return container_id

    def commit_container(self, container_id, image=None, message=None):
        """
        create image from provided container

        :param container_id: str
        :param image: ImageName
        :param message: str
        :return: image_id
        """
        logger.info("committing container '%s'", container_id)
        logger.debug("container_id = '%s', image = '%s', message = '%s'",
                     container_id, image, message)
        tag = None
        if image:
            tag = image.tag
            image = image.to_str(tag=False)
        response = self.d.commit(container_id, repository=image, tag=tag, message=message)
        logger.debug("response = '%s'", response)
        try:
            return response['Id']
        except KeyError:
            logger.error("ID missing from commit response")
            raise RuntimeError("ID missing from commit response")

    def get_image_info_by_image_id(self, image_id):
        """
        using `docker images`, provide information about an image

        :param image_id: str, hash of image to get info
        :return: str or None
        """
        logger.info("getting info about provided image specified by image_id '%s'", image_id)
        logger.debug("image_id = '%s'", image_id)
        # returns list of
        # {u'Created': 1414577076,
        #  u'Id': u'3ab9a7ed8a169ab89b09fb3e12a14a390d3c662703b65b4541c0c7bde0ee97eb',
        #  u'ParentId': u'a79ad4dac406fcf85b9c7315fe08de5b620c1f7a12f45c8185c843f4b4a49c4e',
        #  u'RepoTags': [u'buildroot-fedora:latest'],
        #  u'Size': 0,
        #  u'VirtualSize': 856564160}
        images = self.d.images()
        try:
            image_dict = [i for i in images if i['Id'] == image_id][0]
        except IndexError:
            logger.info("image not found")
            return None
        else:
            return image_dict

    def get_image_info_by_image_name(self, image, exact_tag=True):
        """
        using `docker images`, provide information about an image

        :param image: ImageName, name of image
        :param exact_tag: bool, if false then return info for all images of the
                          given name regardless what their tag is
        :return: list of dicts
        """
        logger.info("getting info about provided image specified by name '%s'", image)
        logger.debug("image_name = '%s'", image)

        # returns list of
        # {u'Created': 1414577076,
        #  u'Id': u'3ab9a7ed8a169ab89b09fb3e12a14a390d3c662703b65b4541c0c7bde0ee97eb',
        #  u'ParentId': u'a79ad4dac406fcf85b9c7315fe08de5b620c1f7a12f45c8185c843f4b4a49c4e',
        #  u'RepoTags': [u'buildroot-fedora:latest'],
        #  u'Size': 0,
        #  u'VirtualSize': 856564160}
        images = self.d.images(name=image.to_str(tag=False))
        if exact_tag:
            # tag is specified, we are looking for the exact image
            for found_image in images:
                if image.to_str(explicit_tag=True) in found_image['RepoTags']:
                    logger.debug("image '%s' found", image)
                    return [found_image]
            images = []  # image not found

        logger.debug("%d matching images found", len(images))
        return images

    def pull_image(self, image, insecure=False):
        """
        pull provided image from registry

        :param image_name: ImageName, image to pull
        :param insecure: bool, allow connecting to registry over plain http
        :return: str, image (reg.om/img:v1)
        """
        logger.info("pulling image '%s' from registry", image)
        logger.debug("image = '%s', insecure = '%s'", image, insecure)
        try:
            logs_gen = self.d.pull(image.to_str(tag=False), tag=image.tag, insecure_registry=insecure, stream=True)
        except TypeError:
            # because changing api is fun
            logs_gen = self.d.pull(image.to_str(tag=False), tag=image.tag, stream=True)
        command_result = wait_for_command(logs_gen)
        self.last_logs = command_result.logs
        return image.to_str()

    def tag_image(self, image, target_image, force=False):
        """
        tag provided image with specified image_name, registry and tag

        :param image: str or ImageName, image to tag
        :param target_image: ImageName, new name for the image
        :param force: bool, force tag the image?
        :return: str, image (reg.om/img:v1)
        """
        logger.info("tagging image '%s' as '%s'", image, target_image)
        logger.debug("image = '%s', target_image_name = '%s'", image, target_image)
        if not isinstance(image, ImageName):
            image = ImageName.parse(image)

        if image != target_image:
            response = self.d.tag(
                image.to_str(),
                target_image.to_str(tag=False),
                tag=target_image.tag,
                force=force)  # returns True/False
            if not response:
                logger.error("failed to tag image")
                raise RuntimeError("Failed to tag image '%s': target_image = '%s'" %
                        image.to_str(), target_image)
        else:
            logger.debug('image already tagged correctly, nothing to do')
        return target_image.to_str()  # this will be the proper name, not just repo/img

    def push_image(self, image, insecure=False):
        """
        push provided image to registry

        :param image: ImageName
        :param insecure: bool, allow connecting to registry over plain http
        :return: str, logs from push
        """
        logger.info("pushing image '%s'", image)
        logger.debug("image: '%s', insecure: '%s'", image, insecure)
        try:
            # push returns string composed of newline separated jsons; exactly what 'docker push' outputs
            logs = self.d.push(image.to_str(tag=False), tag=image.tag, insecure_registry=insecure, stream=True)
        except TypeError:
            # because changing api is fun
            logs = self.d.push(image.to_str(tag=False), tag=image.tag, stream=True)

        command_result = wait_for_command(logs)
        self.last_logs = command_result.logs
        if command_result.is_failed():
            raise RuntimeError("Failed to push image %s" % image)
        return command_result.parsed_logs

    def tag_and_push_image(self, image, target_image, insecure=False, force=False):
        """
        tag provided image and push it to registry

        :param image: str or ImageName, image id or name
        :param target_image: ImageName, img
        :param insecure: bool, allow connecting to registry over plain http
        :param force: bool, force the tag?
        :return: str, image (reg.com/img:v1)
        """
        logger.info("tagging and pushing image '%s' as '%s'", image, target_image)
        logger.debug("image = '%s', target_image = '%s'", image, target_image)
        self.tag_image(image, target_image, force=force)
        return self.push_image(target_image, insecure=insecure)

    def inspect_image(self, image_id):
        """
        return detailed metadata about provided image (see 'man docker-inspect')

        :param image_id: str or ImageName, id or name of the image
        :return: dict
        """
        logger.info("inspecting image '%s'", image_id)
        logger.debug("image_id = '%s'", image_id)
        if isinstance(image_id, ImageName):
            image_id = image_id.to_str()
        image_metadata = self.d.inspect_image(image_id)
        return image_metadata

    def remove_image(self, image_id, force=False, noprune=False):
        """
        remove provided image from filesystem

        :param image_id: str or ImageName
        :param noprune: bool, keep untagged parents?
        :param force: bool, force remove -- just trash it no matter what
        :return: None
        """
        logger.info("removing image '%s' from filesystem", image_id)
        logger.debug("image_id = '%s'", image_id)
        if isinstance(image_id, ImageName):
            image_id = image_id.to_str()
        self.d.remove_image(image_id, force=force, noprune=noprune)  # returns None

    def remove_container(self, container_id, force=False):
        """
        remove provided container from filesystem

        :param container_id: str
        :param force: bool, remove forcefully?
        :return: None
        """
        logger.info("removing container '%s' from filesystem", container_id)
        logger.debug("container_id = '%s'", container_id)
        self.d.remove_container(container_id, force=force)  # returns None

    def logs(self, container_id, stderr=True, stream=True):
        """
        acquire output (stdout, stderr) from provided container

        :param container_id: str
        :param stderr: True, False
        :param stream: if True, return as generator
        :return: either generator, or list of strings
        """
        logger.info("getting stdout of container '%s'", container_id)
        logger.debug("container_id = '%s', stream = '%s'", container_id, stream)
        # returns bytes
        response = self.d.logs(container_id, stdout=True, stderr=stderr, stream=stream)
        if not stream:
            response = response.decode("utf-8")  # py2 & 3 compat
            response = [line for line in response.split('\n') if line]
        return response

    def wait(self, container_id):
        """
        wait for container to finish the job (may run infinitely)

        :param container_id: str
        :return: int, exit code
        """
        logger.info("waiting for container '%s' to finish", container_id)
        logger.debug("container = '%s'", container_id)
        response = self.d.wait(container_id)  # returns exit code as int
        logger.debug("container finished with exit code %s", response)
        return response

    def image_exists(self, image_id):
        """
        does provided image exists?

        :param image_id: str or ImageName
        :return: True if exists, False if not
        """
        logger.info("checking whether image '%s' exists", image_id)
        logger.debug("image_id = '%s'", image_id)
        try:
            response = self.d.inspect_image(image_id)
        except APIError as ex:
            logger.warning(repr(ex))
            response = False
        else:
            response = response is not None
        logger.debug("image exists: %s", response)
        return response

    def get_info(self):
        """
        get info about used docker environment

        :return: dict, json output of `docker info`
        """
        return self.d.info()

    def get_version(self):
        """
        get version of used docker environment

        :return: dict, json output of `docker version`
        """
        return self.d.version()



"""
Copyright (c) 2015 Red Hat, Inc
All rights reserved.

This software may be modified and distributed under the terms
of the BSD license. See the LICENSE file for details.
"""

from __future__ import unicode_literals

from contextlib import contextmanager
from copy import deepcopy
import os
import re
import subprocess
from shutil import rmtree
from tempfile import mkdtemp

from atomic_reactor.plugin import PreBuildPlugin
from atomic_reactor.source import GitSource
from atomic_reactor.plugins.pre_check_and_set_rebuild import is_rebuild
from atomic_reactor.util import get_preferred_label_key
from dockerfile_parse import DockerfileParser


class GitRepo(object):
    """
    In an ideal world, pygit2 would be everywhere we'd need it to be,
    and we could just use that. Meanwhile, in this world, here's a
    wrapper around the git binary.
    """

    def __init__(self, workdir, log, cmd='/usr/bin/git'):
        self.workdir = workdir
        self.log = log
        self.cmd = cmd
        self.tmpdir = mkdtemp()
        self.git_wrapper = os.path.join(self.tmpdir, 'git-wrapper.sh')
        with open(self.git_wrapper, mode='wt') as gitfp:
            gitfp.write('#!/bin/sh\n'
                        'exec /usr/bin/ssh -o StrictHostKeyChecking=no "$@"\n')

        os.chmod(self.git_wrapper, 0o755)

    def __enter__(self):
        return self

    def __exit__(self, exc, value, tb):
        rmtree(self.tmpdir)

    def git(self, args):
        """
        Run git with arguments.

        :param args: list, argument strings
        :return: str, output (including stderr)
        """

        argv = [self.cmd] + args
        env = deepcopy(os.environ)
        env['GIT_SSH'] = self.git_wrapper
        self.log.debug("executing command: %r", argv)
        try:
            with open('/dev/null', 'r+') as devnull:
                output = subprocess.check_output(argv,
                                                 stdin=devnull,
                                                 stderr=subprocess.STDOUT,
                                                 cwd=self.workdir,
                                                 env=env)
        except subprocess.CalledProcessError as ex:
            self.log.debug("command failed, exit code %s", ex.returncode)
            self.log.debug("output: %r", ex.output)
            raise

        output = output.decode().rstrip()
        self.log.debug("git output: %r", output)
        return output


class BumpReleasePlugin(PreBuildPlugin):
    """Git branch management plugin

    For rebuilds, push a new commit incrementing the Release label.

    For initial builds, verify the branch is at the specified commit
    hash.

    When this plugin is configured by osbs-client, the Build's source
    git ref is actually the branch (from --git-branch), not the
    original SHA-1. The SHA-1 specified by --git-commit is stored in
    the configuration for this plugin.

    Some developers will want to specify the release as an ENV
    variable and reference it in the LABEL. This is supported for
    simple situations in which the $VAR reference is at the very
    beginning of the label. In this case, the ENV variable will be
    modified instead of the LABEL.

    Example configuration:

    {
      "name": "bump_release",
      "args": {
        "git_ref": "12345678....",
        "author_name": "OSBS Build System",
        "author_email": "root@example.com"
      }
    }

    Additional optional arguments:
    - committer_name
    - committer_email
    - commit_message
    - push_url

    """

    key = "bump_release"
    is_allowed_to_fail = False  # We really want to stop the process

    def __init__(self, tasker, workflow,
                 git_ref,
                 author_name, author_email,
                 committer_name=None, committer_email=None,
                 commit_message=None,
                 push_url=None):
        """
        constructor

        :param tasker: DockerTasker instance
        :param workflow: DockerBuildWorkflow instance
        :param git_ref: str, commit hash expected on first build
        :param author_name: str, name to use for git commits
        :param author_email: str, email address for git commits
        :param committer_name: str, name to use for git commits (else author's)
        :param committer_email: str, email address for git commits (else
                                     author's)
        :param commit_message: str, git commit message
        :param push_url: str, URL for push
        """
        # call parent constructor
        super(BumpReleasePlugin, self).__init__(tasker, workflow)
        self.git_ref = git_ref
        self.author_name = author_name
        self.author_email = author_email
        self.committer_name = committer_name or author_name
        self.committer_email = committer_email or author_email
        self.push_url = push_url
        self.commit_message = (commit_message or
                               "Bumped release for automated rebuild")

    @contextmanager
    def no_env_replace(self, parser):
        """
        Context manager for temporarily disabling env_replace
        """
        old_val = parser.env_replace
        parser.env_replace = False
        try:
            yield
        finally:
            parser.env_replace = old_val

    def find_current_release(self, parser, label_key):
        """
        Find the attribute set and key name containing the current release.

        This is either the set of labels and the `label_key` we were
        given, or the set of environment variables and the name of the
        environment variable holding the value to change.

        :param parser: DockerfileParser instance
        :param label_key: str, name of release label
        :returns: 2-tuple, attribute set and key name
        """

        attrs = parser.labels
        key = label_key
        value_subst = attrs[key]

        with self.no_env_replace(parser):
            value_nosubst = parser.labels[key]

        # Is the real value stored in an environment variable?
        if value_subst != value_nosubst and value_nosubst.startswith('$'):
            # Yes, but which one?

            # Braced form: a dollar sign, an open brace, the group
            # we're interested in (letters numbers, underscores),
            # followed by a closing brace.
            braced_re = r"\$\{([A-Za-z0-9_]+)\}"

            # Unbraced form: a dollar sign, followed by the group
            # we're interested in (letters, numbers, underscores).
            unbraced_re = r"\$([A-Za-z0-9_]+)"

            match = None
            for pattern in [braced_re, unbraced_re]:
                match = re.match(pattern, value_nosubst)
                self.log.debug("Match %r against %r: %s",
                               value_nosubst, pattern, bool(match))
                if match:
                    break

            if match:
                attrs = parser.envs
                key = match.groups()[0]

        return attrs, key

    @staticmethod
    def get_next_release(current_release):
        """
        Calculate the incremented value.

        :param current_release: str, current value
        :return: str, incremented value
        """

        try:
            return str(int(current_release) + 1)
        except ValueError:
            isdigit = type(current_release).isdigit
            first_nondigit = [isdigit(x) for x in current_release].index(False)
            next_int = str(int(current_release[:first_nondigit]) + 1)
            return next_int + current_release[first_nondigit:]

    def bump(self, repo, remote):
        """
        Push a commit with an incremented release value.

        :param repo: str, GitRepo instance
        :param remote: str, git remote to push to
        """

        # Set up configuration
        repo.git(['config', 'push.default', 'simple'])
        repo.git(['config', 'user.email', self.committer_email])
        repo.git(['config', 'user.name', self.committer_name])
        if self.push_url:
            repo.git(['remote', 'set-url', '--push', remote, self.push_url])

        # Bump the Release label
        df_path = self.workflow.builder.df_path
        parser = DockerfileParser(df_path)
        label_key = get_preferred_label_key(parser.labels, 'release')

        attrs, key = self.find_current_release(parser, label_key)
        next_release = self.get_next_release(attrs[key])
        self.log.info("New Release: %s", next_release)
        attrs[key] = next_release  # this modifies the file

        # Stage it
        repo.git(['add', os.path.basename(df_path)])

        # Commit the change
        repo.git(['commit',
                  '--author={name} <{email}>'.format(name=self.author_name,
                                                     email=self.author_email),
                  '--message={message}'.format(message=self.commit_message)])

        # Push it
        self.log.info("Pushing to git repository")
        repo.git(['push', remote])

    def verify_branch(self, branch, branch_sha):
        """
        Raise exception if the branch is not at the correct commit.

        :param branch: str, branch name
        :param branch_sha: str, commit hash expected
        """

        if branch_sha != self.git_ref:
            self.log.error("Branch '%s' is at commit %s (expected %s)",
                           branch, branch_sha, self.git_ref)
            raise RuntimeError("Not at expected commit")

        self.log.info("Branch '%s' is at expected commit (%s)",
                      branch, self.git_ref)

    def run(self):
        """
        run the plugin
        """

        if self.workflow.build_process_failed:
            self.log.info("Build already failed, not incrementing release")
            return

        # Ensure we can use the git repository already checked out for us
        source = self.workflow.source
        assert isinstance(source, GitSource)
        with GitRepo(source.get(), self.log) as repo:
            # Note: when this plugin is configured by osbs-client,
            # source.git_commit (the Build's source git ref) comes from
            # --git-branch not --git-commit. The value from --git-commit
            # went into our self.git_ref.
            branch = source.git_commit
            try:
                branch_sha = repo.git(['rev-parse', branch])
            except subprocess.CalledProcessError:
                self.log.error("Branch '%s' not found in git repo",
                               source.git_commit)
                raise RuntimeError("Branch '%s' not found" % branch)

            # We checked out the right branch
            assert repo.git(['rev-parse', 'HEAD']) == branch_sha

            # We haven't reset it to an earlier commit
            remote = repo.git(['config', '--get',
                               'branch.{branch}.remote'.format(branch=branch)])
            upstream = '{remote}/{branch}'.format(remote=remote, branch=branch)
            upstream_sha = repo.git(['rev-parse', upstream])
            assert branch_sha == upstream_sha

            if is_rebuild(self.workflow):
                self.log.info("Incrementing release label")
                self.bump(repo, remote)
            else:
                self.log.info("Verifying branch is at specified commit")
                self.verify_branch(branch, branch_sha)



"""
Copyright (c) 2015 Red Hat, Inc
All rights reserved.

This software may be modified and distributed under the terms
of the BSD license. See the LICENSE file for details.


Add arbitrary yum repo, specified by URL of repo file, to a list of
repos which should be injected into built image by the inject_yum_repo
plugin.

This plugin has to run _BEFORE_ the inject_yum_repo plugin, which
actually places the repo file in the build environment.

Example configuration to add content of repo file at URL:

{
    "name": "add_yum_repo_by_url",
    "args": {
        "repourls": ["http://example.com/myrepo/myrepo.repo"]
    }
}

"""
from atomic_reactor.constants import YUM_REPOS_DIR
from atomic_reactor.plugin import PreBuildPlugin
import os
import os.path
import requests

try:
    # py2
    from urlparse import unquote, urlsplit
except ImportError:
    # py3
    from urllib.parse import unquote, urlsplit


class YumRepo(object):
    def __init__(self, repourl, dst_repos_dir=YUM_REPOS_DIR):
        self.repourl = repourl
        self.dst_repos_dir = dst_repos_dir
        self.content = None

    @property
    def filename(self):
        urlpath = unquote(urlsplit(self.repourl, allow_fragments=False).path)
        return os.path.basename(urlpath)

    @property
    def dst_filename(self):
        return os.path.join(self.dst_repos_dir, self.filename)

    def fetch(self):
        response = requests.get(self.repourl)
        response.raise_for_status()
        self.content = response.content


class AddYumRepoByUrlPlugin(PreBuildPlugin):
    key = "add_yum_repo_by_url"
    is_allowed_to_fail = False

    def __init__(self, tasker, workflow, repourls):
        """
        constructor

        :param tasker: DockerTasker instance
        :param workflow: DockerBuildWorkflow instance
        :param repourls: list of str, URLs to the repo files
        """
        # call parent constructor
        super(AddYumRepoByUrlPlugin, self).__init__(tasker, workflow)
        self.repourls = repourls

    def run(self):
        """
        run the plugin
        """
        if self.repourls:
            for repourl in self.repourls:
                yumrepo = YumRepo(repourl)
                yumrepo.fetch()
                self.log.info("fetched repo from '%s'", yumrepo.repourl)
                self.workflow.files[yumrepo.dst_filename] = yumrepo.content
                self.log.debug("saving repo '%s', length %d", yumrepo.dst_filename, len(yumrepo.content))



"""
Copyright (c) 2015 Red Hat, Inc
All rights reserved.

This software may be modified and distributed under the terms
of the BSD license. See the LICENSE file for details.
"""

from __future__ import print_function, unicode_literals
import os
import json

from atomic_reactor.plugins.input_osv3 import OSv3InputPlugin

import pytest
from flexmock import flexmock

from tests.constants import LOCALHOST_REGISTRY

@pytest.mark.parametrize('prebuild_json,expected_json', [
    ([],
     [{ 'name': 'pull_base_image' }]
    ),
    ([{ 'name': 'pull_base_image' }],
     [{ 'name': 'pull_base_image' }]
    ),
    ([{ 'name': 'pull_base_image', 'args': { 'a': 'b' }}],
     [{ 'name': 'pull_base_image', 'args': { 'a': 'b' }}]
    ),
    ([{ 'name': 'change_source_registry' }],
     [{ 'name': 'pull_base_image' }]
    ),
    ([{ 'name': 'change_source_registry',
        'args': { 'registry_uri': 'localhost:666', 'insecure_registry': True }}],
     [{ 'name': 'pull_base_image',
        'args': { 'parent_registry': 'localhost:666', 'parent_registry_insecure': True }}]
    ),
    ([{ 'name': 'change_source_registry' }, { 'name': 'pull_base_image', 'args': { 'a': 'b' }}],
     [{ 'name': 'pull_base_image', 'args': { 'a': 'b' }}]
    ),
])
def test_prebuild_plugins_rewrite(prebuild_json, expected_json):
    plugins_json = {
        'prebuild_plugins': prebuild_json,
    }

    mock_env = {
        'BUILD': '{}',
        'SOURCE_URI': 'https://github.com/foo/bar.git',
        'SOURCE_REF': 'master',
        'OUTPUT_IMAGE': 'asdf:fdsa',
        'OUTPUT_REGISTRY': 'localhost:5000',
        'ATOMIC_REACTOR_PLUGINS': json.dumps(plugins_json),
    }
    flexmock(os, environ=mock_env)

    plugin = OSv3InputPlugin()
    assert plugin.run()['prebuild_plugins'] == expected_json


@pytest.mark.parametrize('output_registry,postbuild_json,expected_json', [
    (LOCALHOST_REGISTRY,
     [],
     [{ 'name': 'tag_and_push', 'args': { 'registries': { LOCALHOST_REGISTRY: { 'insecure': True }}}}]
    ),
    (LOCALHOST_REGISTRY,
     [{ 'name': 'tag_and_push' }],
     [{ 'name': 'tag_and_push', 'args': { 'registries': { LOCALHOST_REGISTRY: { 'insecure': True }}}}]
    ),
    (LOCALHOST_REGISTRY,
     [{ 'name': 'tag_and_push', 'args': { 'registries': { 'foo': { 'insecure': True }}}}],
     [{ 'name': 'tag_and_push', 'args': { 'registries': { 'foo': { 'insecure': True }}}}]
    ),
    (None,
     [{ 'name': 'tag_and_push', 'args': { 'registries': { 'foo': { 'insecure': True }}}}],
     [{ 'name': 'tag_and_push', 'args': { 'registries': { 'foo': { 'insecure': True }}}}]
    ),
    (None,
     [],
     []
    ),
])
def test_postbuild_plugins_rewrite(output_registry, postbuild_json, expected_json):
    plugins_json = {
        'postbuild_plugins': postbuild_json,
    }

    mock_env = {
        'BUILD': '{}',
        'SOURCE_URI': 'https://github.com/foo/bar.git',
        'SOURCE_REF': 'master',
        'OUTPUT_IMAGE': 'asdf:fdsa',
        'OUTPUT_REGISTRY': output_registry,
        'ATOMIC_REACTOR_PLUGINS': json.dumps(plugins_json),
    }
    flexmock(os, environ=mock_env)

    plugin = OSv3InputPlugin()
    assert plugin.run()['postbuild_plugins'] == expected_json


def test_doesnt_fail_if_no_plugins():
    mock_env = {
        'BUILD': '{}',
        'SOURCE_URI': 'https://github.com/foo/bar.git',
        'SOURCE_REF': 'master',
        'OUTPUT_IMAGE': 'asdf:fdsa',
        'OUTPUT_REGISTRY': 'localhost:5000',
        'ATOMIC_REACTOR_PLUGINS': '{}',
    }
    flexmock(os, environ=mock_env)

    plugin = OSv3InputPlugin()
    assert plugin.run()['prebuild_plugins'] == [{'name': 'pull_base_image'}]


@pytest.mark.parametrize('build, expected', [
    ('{"metadata": {"selfLink": "/foo/bar"}}', '/foo/bar'),
    ('{"metadata": {}}', None),
    ('{}', None),
])
def test_sets_selflink(build, expected):
    mock_env = {
        'BUILD': build,
        'SOURCE_URI': 'https://github.com/foo/bar.git',
        'SOURCE_REF': 'master',
        'OUTPUT_IMAGE': 'asdf:fdsa',
        'OUTPUT_REGISTRY': 'localhost:5000',
        'ATOMIC_REACTOR_PLUGINS': '{}',
    }
    flexmock(os, environ=mock_env)

    plugin = OSv3InputPlugin()
    assert plugin.run()['openshift_build_selflink'] == expected


@pytest.mark.parametrize('plugins_variable', ['ATOMIC_REACTOR_PLUGINS', 'DOCK_PLUGINS'])
def test_plugins_variable(plugins_variable):
    plugins_json = {
        'postbuild_plugins': [],
    }

    mock_env = {
        'BUILD': '{}',
        'SOURCE_URI': 'https://github.com/foo/bar.git',
        'SOURCE_REF': 'master',
        'OUTPUT_IMAGE': 'asdf:fdsa',
        'OUTPUT_REGISTRY': 'localhost:5000',
        plugins_variable: json.dumps(plugins_json),
    }
    flexmock(os, environ=mock_env)

    plugin = OSv3InputPlugin()
    assert plugin.run()['postbuild_plugins'] is not None



"""
Copyright (c) 2015 Red Hat, Inc
All rights reserved.

This software may be modified and distributed under the terms
of the BSD license. See the LICENSE file for details.

Push built image to pulp registry

Several authentication schemes are possible, including
username+password and key/certificate via sourceSecret.

However, the recommended scheme (since Origin 1.0.6) is to store a
key and certificate in a secret which the builder service account is
allowed to mount:

$ oc secrets new pulp ./pulp.key ./pulp.cer
secrets/pulp
$ oc secrets add serviceaccount/builder secret/pulp --for=mount

In the BuildConfig for atomic-reactor, specify the secret in the
strategy's 'secrets' array, specifying a mount path:

"secrets": [{
  "secretSource": {
    "name": "pulp"
  },
  "mountPath": "/var/run/secrets/pulp"
}]

In the configuration for this plugin, specify the same path for
pulp_secret_path:

"pulp_push": {
  "pulp_registry_name": ...,
  ...
  "pulp_secret_path": "/var/run/secrets/pulp"
}
"""

from __future__ import print_function, unicode_literals
from dockpulp import setup_logger
from atomic_reactor import set_logging

from atomic_reactor.plugin import PostBuildPlugin
from atomic_reactor.util import ImageName

import dockpulp
import dockpulp.imgutils

import os
import re
import tempfile
from collections import namedtuple


# let's silence warnings from dockpulp: there is one warning for every request
# which may result in tenths of messages: very annoying
# with "module", it just prints one warning -- this should balance security and UX
import warnings
warnings.filterwarnings("module")


PulpRepo = namedtuple('PulpRepo', ['registry_id', 'tags'])

class PulpUploader(object):
    CER = 'pulp.cer'
    KEY = 'pulp.key'

    def __init__(self, workflow, pulp_instance, filename, log, pulp_secret_path=None, username=None,
                 password=None):
        self.workflow = workflow
        self.pulp_instance = pulp_instance
        self.filename = filename
        self.pulp_secret_path = pulp_secret_path
        self.log = log
        # U/N & password has bigger prio than secret cert
        self.username = username
        self.password = password

        set_logging("dockpulp")

    def _check_file(self):
        # Sanity-check image
        metadata = dockpulp.imgutils.get_metadata(self.filename)
        vers = dockpulp.imgutils.get_versions(metadata)
        for _, version in vers.items():
            verparts = version.split('.')
            major = int(verparts[0])
            if major < 1:
                minor = 0
                if len(verparts) > 1:
                    minor = int(verparts[1])
                if minor < 10:
                    raise RuntimeError('An image layer uses an unsupported '
                                       'version of docker (%s)' % version)

        r_chk = dockpulp.imgutils.check_repo(self.filename)
        if r_chk == 1:
            raise RuntimeError('Image is missing a /repositories file')
        elif r_chk == 2:
            raise RuntimeError('Pulp demands exactly 1 repo in /repositories')
        elif r_chk == 3:
            raise RuntimeError('/repositories references external images')

    def _set_auth(self, p):
        # The pulp.cer and pulp.key values must be set in a
        # 'Secret'-type resource and mounted somewhere we can get at them.
        if self.username and self.password:
            p.login(self.username, self.password)
        elif self.pulp_secret_path or 'SOURCE_SECRET_PATH' in os.environ:
            if self.pulp_secret_path is not None:
                path = self.pulp_secret_path
                self.log.info("using configured path %s for secrets", path)
            else:
                path = os.environ["SOURCE_SECRET_PATH"]
                self.log.info("SOURCE_SECRET_PATH=%s from environment", path)

            # Work out the pathnames for the certificate/key pair.
            cer = os.path.join(path, self.CER)
            key = os.path.join(path, self.KEY)

            if not os.path.exists(cer):
                raise RuntimeError("Certificate does not exist.")
            if not os.path.exists(key):
                raise RuntimeError("Key does not exist.")

            # Tell dockpulp.
            p.set_certs(cer, key)

    def _create_missing_repos(self, p, pulp_repos, repo_prefix):
        repos = pulp_repos.keys()
        found_repos = p.getRepos(repos, fields=["id"])
        found_repo_ids = [repo["id"] for repo in found_repos]

        missing_repos = set(repos) - set(found_repo_ids)
        self.log.info("Missing repos: %s" % ", ".join(missing_repos))
        for repo in missing_repos:
            p.createRepo(repo, "/pulp/docker/%s" % repo,
                         registry_id=pulp_repos[repo].registry_id,
                         prefix_with=repo_prefix)

    def _get_tar_metadata(self, tarfile):
        metadata = dockpulp.imgutils.get_metadata(tarfile)
        pulp_md = dockpulp.imgutils.get_metadata_pulp(metadata)
        layers = pulp_md.keys()
        top_layer = dockpulp.imgutils.get_top_layer(pulp_md)

        return top_layer, layers

    def push_tarball_to_pulp(self, image_names, repo_prefix="redhat-"):
        self.log.info("checking image before upload")
        self._check_file()

        p = dockpulp.Pulp(env=self.pulp_instance)
        self._set_auth(p)

        # pulp_repos is mapping from repo-ids to registry-ids and tags
        # which should be applied to those repos, expected structure:
        # {
        #    "my-image": PulpRepo(registry_id="nick/my-image", tags=["v1", "latest"])
        #    ...
        # }
        pulp_repos = {}
        for image in image_names:
            repo_id = image.pulp_repo
            tag = image.tag if image.tag else 'latest'
            if repo_prefix:
                repo_id = repo_prefix + repo_id

            if repo_id in pulp_repos:
                pulp_repos[repo_id].tags.append(tag)
            else:
                pulp_repos[repo_id] = PulpRepo(
                    registry_id=image.to_str(registry=False, tag=False),
                    tags=[tag]
                )

        top_layer, layers = self._get_tar_metadata(self.filename)
        self.log.info("pulp_repos = %s", pulp_repos)
        self._create_missing_repos(p, pulp_repos, repo_prefix)

        p.upload(self.filename)

        for repo_id, pulp_repo in pulp_repos.items():
            for layer in layers:
                p.copy(repo_id, layer)
            p.updateRepo(repo_id, {"tag": "%s:%s" % (",".join(pulp_repo.tags),
                                                     top_layer)})

        task_ids = p.crane(pulp_repos.keys(), wait=True)
        self.log.info("waiting for repos to be published to crane, tasks: %s",
                      ", ".join(map(str, task_ids)))
        p.watch_tasks(task_ids)

        # Store the registry URI in the push configuration

        # We only want the hostname[:port]
        pulp_registry = re.sub(r'^https?://([^/]*)/?.*',
                               lambda m: m.groups()[0],
                               p.registry)

        self.workflow.push_conf.add_pulp_registry(self.pulp_instance,
                                                  pulp_registry)

        # Return the set of qualified repo names for this image
        return [ImageName(registry=pulp_registry, repo=repodata.registry_id, tag=tag)
                for dummy_repo, repodata in pulp_repos.items()
                for tag in repodata.tags]


class PulpPushPlugin(PostBuildPlugin):
    key = "pulp_push"
    is_allowed_to_fail = False

    def __init__(self, tasker, workflow, pulp_registry_name, load_squashed_image=None,
                 load_exported_image=None, image_names=None, pulp_secret_path=None,
                 username=None, password=None, dockpulp_loglevel=None):
        """
        constructor

        :param tasker: DockerTasker instance
        :param workflow: DockerBuildWorkflow instance
        :param pulp_registry_name: str, name of pulp registry to use, specified in /etc/dockpulp.conf
        :param load_squashed_image: obsolete name for load_exported_image, please don't use
        :param load_exported_image: bool, use exported tar instead of image from Docker
        :param image_names: list of additional image names
        :param pulp_secret_path: path to pulp.cer and pulp.key; $SOURCE_SECRET_PATH otherwise
        :param username: pulp username, used in preference to certificate and key
        :param password: pulp password, used in preference to certificate and key
        """
        # call parent constructor
        super(PulpPushPlugin, self).__init__(tasker, workflow)
        self.pulp_registry_name = pulp_registry_name
        self.image_names = image_names
        if load_squashed_image is not None and load_exported_image is not None and \
                (load_squashed_image != load_exported_image):
            raise RuntimeError(
                'Can\'t use load_squashed_image and load_exported_image with different values')
        if load_squashed_image is not None:
            self.log.warning(
                'load_squashed_image argument is obsolete and will be removed in a future version;'
                'please use load_exported_image instead')
        self.load_exported_image = load_exported_image or load_squashed_image or False
        self.pulp_secret_path = pulp_secret_path
        self.username = username
        self.password = password

        if dockpulp_loglevel is not None:
            logger = setup_logger(dockpulp.log)
            try:
                logger.setLevel(dockpulp_loglevel)
            except (ValueError, TypeError) as ex:
                self.log.error("Can't set provided log level %r: %r", dockpulp_loglevel, ex)

    def push_tar(self, image_path, image_names=None):
        # Find out how to tag this image.
        self.log.info("image names: %s", [str(image_name) for image_name in image_names])

        # Give that compressed tarball to pulp.
        uploader = PulpUploader(self.workflow, self.pulp_registry_name, image_path, self.log,
                                pulp_secret_path=self.pulp_secret_path, username=self.username,
                                password=self.password)
        return uploader.push_tarball_to_pulp(image_names)

    def run(self):
        image_names = self.workflow.tag_conf.images[:]
        # Add in additional image names, if any
        if self.image_names:
            self.log.info("extending image names: %s", self.image_names)
            image_names += [ImageName.parse(x) for x in self.image_names]

        if self.load_exported_image:
            if len(self.workflow.exported_image_sequence) == 0:
                raise RuntimeError('no exported image to push to pulp')
            crane_repos = self.push_tar(self.workflow.exported_image_sequence[-1].get("path"),
                                        image_names)
        else:
            # Work out image ID
            image = self.workflow.image
            self.log.info("fetching image %s from docker", image)
            with tempfile.NamedTemporaryFile(prefix='docker-image-', suffix='.tar') as image_file:
                image_file.write(self.tasker.d.get_image(image).data)
                crane_repos = self.push_tar(image_file.name, image_names)

        for image_name in crane_repos:
            self.log.info("image available at %s", str(image_name))

        return crane_repos



"""
Copyright (c) 2015 Red Hat, Inc
All rights reserved.

This software may be modified and distributed under the terms
of the BSD license. See the LICENSE file for details.


Reads input from provided path
"""
import json
import os

from atomic_reactor.constants import CONTAINER_BUILD_JSON_PATH

from atomic_reactor.plugin import InputPlugin


class PathInputPlugin(InputPlugin):
    key = "path"

    def __init__(self, path=None, **kwargs):
        """
        constructor
        """
        # call parent constructor
        super(PathInputPlugin, self).__init__(**kwargs)
        self.path = path

    def run(self):
        """
        get json with build config from path
        """
        path = self.path or CONTAINER_BUILD_JSON_PATH
        try:
            with open(path, 'r') as build_cfg_fd:
                build_cfg_json = json.load(build_cfg_fd)
        except ValueError:
            self.log.error("couldn't decode json from file '%s'", path)
            return None
        except IOError:
            self.log.error("couldn't read json from file '%s'", path)
            return None
        else:
            return self.substitute_configuration(build_cfg_json)

    @classmethod
    def is_autousable(cls):
        return os.path.exists(CONTAINER_BUILD_JSON_PATH)



"""
Copyright (c) 2015 Red Hat, Inc
All rights reserved.

This software may be modified and distributed under the terms
of the BSD license. See the LICENSE file for details.


It returns content of dockerfile and therefore displays it in results.
"""
from dockerfile_parse import DockerfileParser
from atomic_reactor.plugin import PreBuildPlugin


class CpDockerfilePlugin(PreBuildPlugin):
    key = "dockerfile_content"

    def __init__(self, tasker, workflow):
        """
        constructor

        :param tasker: DockerTasker instance
        :param workflow: DockerBuildWorkflow instance
        :return:
        """
        # call parent constructor
        super(CpDockerfilePlugin, self).__init__(tasker, workflow)

    def run(self):
        """
        try open dockerfile, output an error if there is one
        """
        try:
            return DockerfileParser(self.workflow.builder.df_path).content
        except (IOError, OSError) as ex:
            return "Couldn't retrieve dockerfile: %r" % ex



"""
Copyright (c) 2015 Red Hat, Inc
All rights reserved.

This software may be modified and distributed under the terms
of the BSD license. See the LICENSE file for details.
"""

from __future__ import unicode_literals

try:
    import pygit2
except ImportError:
    pygit2 = None

import pytest
from atomic_reactor.core import DockerTasker
from atomic_reactor.inner import DockerBuildWorkflow
from atomic_reactor.plugin import PreBuildPluginsRunner, PluginFailedException
from atomic_reactor.plugins.\
    pre_check_and_set_rebuild import CheckAndSetRebuildPlugin
from atomic_reactor.plugins.pre_bump_release import BumpReleasePlugin
from atomic_reactor.source import GitSource
from atomic_reactor.util import ImageName
from tests.constants import SOURCE, MOCK
if MOCK:
    from tests.docker_mock import mock_docker

from copy import deepcopy
from dockerfile_parse import DockerfileParser
import os
import shutil
import tempfile
import subprocess

from flexmock import flexmock


BRANCH = 'branch'


class DFWithRelease(object):
    def __init__(self, label=None, lines=None):
        self.path = tempfile.mkdtemp()
        if label is None and lines is None:
            label = "LABEL Release 1"

        if lines is None:
            lines = ['FROM baseimage\n',
                     '{0}\n'.format(label)]

        self.lines = lines

    def __fini__(self):
        shutil.rmtree(self.path)

    def __enter__(self):
        repo = pygit2.init_repository(self.path)
        repo.remotes.create('origin', '/dev/null')

        # Set up branch 'master'
        filename = 'Dockerfile'
        dockerfile_path = os.path.join(self.path, filename)
        open(dockerfile_path, mode="w+t").close()
        index = repo.index
        index.add(filename)
        author = pygit2.Signature('Test', 'test@example.com')
        committer = pygit2.Signature('Test', 'test@example.com')
        oid = repo.create_commit('HEAD', author, committer,
                                 '', index.write_tree(), [])
        master = repo.head

        # Now set up our branch
        branch = repo.create_branch(BRANCH, repo.get(oid))
        repo.checkout(refname=branch)
        DockerfileParser(dockerfile_path).lines = self.lines

        index = repo.index
        index.add(filename)
        repo.create_commit(branch.name, author, committer,
                           '', index.write_tree(),
                           [repo.head.peel().hex])
        branch.upstream = branch
        return dockerfile_path, repo.head.peel().hex

    def __exit__(self, exc, value, tb):
        pass


class X(object):
    pass


class MockedPopen(subprocess.Popen):
    def __init__(self, cmd, *args, **kwargs):
        # Don't actually push any commits
        if 'push' in cmd:
            cmd = ['/bin/echo', 'push faked']

        # Deal with there being no remote for our branch
        elif cmd[1:] == ['config', '--get', 'branch.%s.remote' % BRANCH]:
            cmd = ['/bin/echo', 'origin']
        elif cmd[1] == 'rev-parse' and cmd[2].startswith('origin/'):
            cmd[2] = cmd[2][len('origin/'):]

        super(MockedPopen, self).__init__(cmd, *args, **kwargs)


def prepare(tmpdir, df_path, git_ref, source=None,
            build_process_failed=False, is_rebuild=True,
            author_name=None, author_email=None,
            commit_message=None, git_commit=None):
    if author_name is None:
        author_name = "OSBS Build System"
    if author_email is None:
        author_email = "root@example.com"

    if MOCK:
        mock_docker()
    tasker = DockerTasker()
    source = deepcopy(SOURCE)
    source['provider_params']['git_commit'] = git_commit or BRANCH
    workflow = DockerBuildWorkflow(source, "test-image")
    setattr(workflow, 'builder', X())
    setattr(workflow.builder, 'image_id', 'asd123')
    setattr(workflow.builder, 'base_image', ImageName(repo='Fedora', tag='22'))
    setattr(workflow.builder, 'source', workflow.source)
    setattr(workflow.builder, 'df_path', df_path)
    setattr(workflow, 'plugin_failed', build_process_failed)
    workflow.prebuild_results[CheckAndSetRebuildPlugin.key] = is_rebuild
    flexmock(GitSource, get=lambda: os.path.dirname(df_path))
    flexmock(subprocess, Popen=MockedPopen)
    args = {
        'git_ref': git_ref,
        'author_name': author_name,
        'author_email': author_email,
        'commit_message': commit_message,
        'push_url': '/',
    }
    for omitted in [x for x in args if args[x] is None]:
        del args[omitted]
    runner = PreBuildPluginsRunner(tasker, workflow,
                                   [
                                       {
                                           'name': BumpReleasePlugin.key,
                                           'args': args,
                                       }
                                   ])
    return workflow, args, runner


@pytest.mark.skipif(pygit2 is None,
                    reason="pygit2 required for this test")
def test_bump_release_failed_build(tmpdir):
    with DFWithRelease() as (df_path, commit):
        workflow, args, runner = prepare(tmpdir, df_path, commit,
                                         build_process_failed=True)
        original_content = open(df_path).readlines()
        runner.run()
        assert open(df_path).readlines() == original_content


@pytest.mark.skipif(pygit2 is None,
                    reason="pygit2 required for this test")
def test_bump_release_not_rebuild_extra_commits(tmpdir):
    with DFWithRelease() as (df_path, commit):
        workflow, args, runner = prepare(tmpdir, df_path, "wrongcommit",
                                         is_rebuild=False)
        with pytest.raises(PluginFailedException):
            runner.run()


@pytest.mark.skipif(pygit2 is None,
                    reason="pygit2 required for this test")
def test_bump_release_not_rebuild_no_extra_commits(tmpdir):
    with DFWithRelease() as (df_path, commit):
        workflow, args, runner = prepare(tmpdir, df_path, commit,
                                         is_rebuild=False)
        runner.run()


@pytest.mark.skipif(pygit2 is None,
                    reason="pygit2 required for this test")
def test_bump_release_branch_not_found(tmpdir):
    with DFWithRelease() as (df_path, commit):
        workflow, args, runner = prepare(tmpdir, df_path, commit,
                                         git_commit='wrong')
        with pytest.raises(PluginFailedException):
            runner.run()


@pytest.mark.skipif(pygit2 is None,
                    reason="pygit2 required for this test")
@pytest.mark.parametrize(('label', 'expected'), [
    # Simple case, no '=' or quotes
    ('LABEL Release 1',
     'LABEL Release 2'),

    # No '=' but quotes
    ('LABEL "Release" "2"',
     'LABEL Release 3'),

    # Deal with another label
    ('LABEL Release 3\nLABEL Name foo',
     'LABEL Release 4'),

    # Simple case, '=' but no quotes
    ('LABEL Release=1',
     'LABEL Release=2'),

    # '=' and quotes
    ('LABEL "Release"="2"',
     'LABEL Release=3'),

    # '=', multiple labels, no quotes
    ('LABEL Name=foo Release=3',
     'LABEL Name=foo Release=4'),

    # '=', multiple labels and quotes
    ('LABEL Name=foo "Release"="4"',
     'LABEL Name=foo Release=5'),

    # Release that's not entirely numeric
    ('LABEL Release=1.1',
     'LABEL Release=2.1'),
])
def test_bump_release_direct(tmpdir, label, expected):
    with DFWithRelease(label=label) as (df_path, commit):
        workflow, args, runner = prepare(tmpdir, df_path, commit,
                                         commit_message='foo')
        runner.run()
        assert open(df_path).readlines()[1].rstrip() == expected
        repo = pygit2.init_repository(os.path.dirname(df_path))

        # We're on the 'branch' branch
        assert repo.head.name == repo.lookup_branch(BRANCH).name

        # and one commit ahead of where we were
        assert repo.head.peel().parents[0].hex == commit

        # Examine metadata for newest commit
        author = repo.head.peel().author
        assert author.name == args['author_name']
        assert author.email == args['author_email']

        committer = repo.head.peel().committer
        assert ('committer_name' not in args or
                committer.name == args['committer_name'])
        assert ('committer_email' not in args or
                committer.email == args['committer_email'])

        if 'commit_message' in args:
            assert repo.head.peel().message.rstrip() == args['commit_message']


@pytest.mark.skipif(pygit2 is None,
                    reason="pygit2 required for this test")
@pytest.mark.parametrize('labelval', [
    # Simple case, no quotes
    '$RELEASE',

    # Double quotes
    '"$RELEASE"',

    # Braces, no quotes
    '${RELEASE}',

    # Braces, double quotes
    '"${RELEASE}"',
])
def test_bump_release_indirect_correct(tmpdir, labelval):
    dflines = ['FROM fedora\n',
               'ENV RELEASE=1\n',
               'LABEL Release={0}\n'.format(labelval)]
    with DFWithRelease(lines=dflines) as (df_path, commit):
        dummy_workflow, dummy_args, runner = prepare(tmpdir, df_path, commit)
        labels_before = DockerfileParser(df_path, env_replace=False).labels

        runner.run()

        parser = DockerfileParser(df_path)
        assert parser.envs['RELEASE'] == '2'

        parser.env_replace = False
        assert parser.labels == labels_before


@pytest.mark.skipif(pygit2 is None,
                    reason="pygit2 required for this test")
@pytest.mark.parametrize('labelval', [
    # Single quotes
    "'$RELEASE'",

    # Braces, single quotes
    "'${RELEASE}'",

    # Escaped, no quotes
    '\\$RELEASE',

    # Escaped, single quotes
    "'\\$RELEASE'",

    # Escaped, double quotes
    '"\\$RELEASE"',

    # Escaped, braces, no quotes
    "\\${RELEASE}",

    # Escaped, braces, single quotes
    "'\\${RELEASE}'",

    # Escaped, braces, double quotes
    '"\\${RELEASE}"',
])
def test_bump_release_indirect_incorrect(tmpdir, labelval):
    dflines = ['FROM fedora\n',
               'ENV RELEASE=1\n',
               'LABEL Release={0}\n'.format(labelval)]
    with DFWithRelease(lines=dflines) as (df_path, commit):
        dummy_workflow, dummy_args, runner = prepare(tmpdir, df_path, commit)

        with pytest.raises(PluginFailedException):
            runner.run()

        assert DockerfileParser(df_path).lines == dflines


@pytest.mark.parametrize(('labelval', 'expected_attr', 'expected_key'), [
    # Direct
    ('1', 'labels', 'Label'),

    # Simple case, no quotes
    ('$ENV', 'envs', 'ENV'),

    # Check word boundaries
    ('$ENV-dev', 'envs', 'ENV'),

    # Double quotes
    ('"$ENV"', 'envs', 'ENV'),

    # Braces, no quotes
    ('${ENV}', 'envs', 'ENV'),

    # Braces, double quotes
    ('"${ENV}"', 'envs', 'ENV'),
])
def test_bump_release_find_current_release(tmpdir,
                                           labelval,
                                           expected_attr,
                                           expected_key):
    dflines = ['FROM fedora\n',
               'ENV ENV=1\n',
               'LABEL Label={0}\n'.format(labelval)]
    df_path = os.path.join(str(tmpdir), 'Dockerfile')
    DockerfileParser(df_path).lines = dflines
    plugin = BumpReleasePlugin(None, None, None, None, None)
    parser = DockerfileParser(df_path)
    attrs, key = plugin.find_current_release(parser, 'Label')
    assert attrs == getattr(parser, expected_attr)
    assert key == expected_key
